"""
[script] section from **snowmobile.toml**, including subsections.
"""
from __future__ import annotations

import re
import itertools
from typing import Dict, List, Tuple, Union, Iterable, Optional

import pandas as pd
import sqlparse
from pydantic import Field

from snowmobile.core.utils import parsing as p
from snowmobile.core import errors

from .base import Base


class Wildcard(Base):
    """[script.patterns.wildcards]"""

    # fmt: off
    char_wc: str = Field(
        default_factory=str, alias="wildcard-character"
    )
    char_sep: str = Field(
        default_factory=str, alias="wildcard-delimiter"
    )
    wc_paragraph: str = Field(
        default_factory=str, alias="denotes-paragraph"
    )
    wc_as_is: str = Field(
        default_factory=str, alias="denotes-no-reformat"
    )
    wc_omit_attr_nm: str = Field(
        default_factory=str, alias="denotes-omit-name-in-output"
    )
    # fmt: on

    def find_first_wc_idx(self, attr_nm: str) -> int:
        """Finds index of the first unescaped wildcard in an attribute name.

        Args:
            attr_nm (str): Attribute name to search through.

        Returns (int):
            Index position of first unescaped wildcard or 0 if one does not exist.

        """
        idx = 0
        for i, c in enumerate(attr_nm):
            if c == self.char_wc and "\\" != attr_nm[i - 1]:
                idx = i
                break
        return idx

    def partition_on_wc(self, attr_nm: str) -> Tuple[str, List[str]]:
        """Parses attribute name into its display name and its wildcards.

        Uses :meth:`Wildcard.find_first_wc_idx()` to determine if **attr_nm**
        contains a valid wildcard.

        Args:
            attr_nm (str): Attribute name to parse.

        Returns (Tuple[str, List[str]]):
            Tuple containing the attribute display name and a list of its
            wildcards if present and an empty list otherwise.

        """
        idx_of_first_unescaped = self.find_first_wc_idx(attr_nm=attr_nm)
        has_valid_wc = idx_of_first_unescaped != 0

        wildcards = []
        if has_valid_wc:
            nm_to_strip = attr_nm[:idx_of_first_unescaped]
            wildcards = attr_nm[idx_of_first_unescaped:].split("_")
        else:
            nm_to_strip = attr_nm

        stripped_nm = nm_to_strip.replace(f"\\{self.char_wc}", self.char_wc)
        return stripped_nm, wildcards


class Reserved(Base):
    """[script.markdown.attributes.reserved]"""

    # fmt: off
    include_by_default: bool = Field(
        default_factory=bool, alias="include-by-default"
    )
    attr_nm: str = Field(
        default_factory=str, alias="attribute-name"
    )
    default_val: str = Field(
        default_factory=str, alias="default-to"
    )
    default_format: str = Field(
        default_factory=str, alias="format"
    )
    tabulate_format: str = Field(
        default_factory=str, alias="tabulate-format"
    )
    # fmt: on
    

class Marker(Base):
    """[script.markdown.attributes.markers]"""

    # fmt: off
    name: str = Field(
        default_factory=str, alias="name"
    )
    group: str = Field(
        default_factory=str, alias="as-group"
    )
    attrs: Dict = Field(
        default_factory=dict, alias="wrap"
    )
    raw: str = Field(
        default_factory=int, alias="raw-text"
    )
    index: int = Field(
        default_factory=int, alias="index"
    )
    # fmt: on

    def __init__(self, **data):
        super().__init__(**data)
        if not self.attrs:
            self.attrs = {
                k: v for k, v in data.items() if k not in self.dict(by_alias=True)
            }

    def add(self, attrs: Dict) -> Marker:
        """Add to existing attributes."""
        for k, v in attrs.items():
            self.attrs[k] = v
        return self

    def split_attrs(self, attrs: Dict) -> Tuple[Dict, Dict]:
        """Splits attributes into user-defined-only and shared with snowmobile.toml.

        Args:
            attrs (Dict):
                Dictionary of parsed arguments.

        Returns (Tuple[Dict, Dict]):
            Tuple of (shared_with_snowmobile_toml_attrs, new_attrs)

        """
        shared = {k: v for k, v in attrs.items() if k in self.attrs}
        new = {k: attrs[k] for k, v in attrs.items() if k not in self.attrs}
        return shared, new

    def update(self, attrs: Dict) -> Marker:
        """Merges parsed attributes with configuration attributes"""
        self.raw = attrs.pop("raw-text")
        shared_attrs, new_attrs = self.split_attrs(attrs=attrs)
        self.attrs.update(shared_attrs)
        self.attrs = {self.group: self.attrs} if self.group else self.attrs
        self.add(attrs=new_attrs)
        return self

    def set_name(self, name: str, overwrite: bool = False) -> Marker:
        """Sets the name attribute."""
        if self.name and not overwrite:
            return self
        self.name = name.replace("_", "")
        return self

    def as_args(self):
        """Returns a dictionary of arguments for :class:`Section`."""
        if self.attrs.get("name"):
            self.name = self.attrs.pop("name")
        if self.attrs.get("marker-name"):
            _ = self.attrs.pop("marker-name")
        return {"h_contents": self.nm(), "parsed": self.attrs, "is_marker": True}
    
    def nm(self):
        """Marker name."""
        return self.name

    def __str__(self):
        return f"Marker('{self.nm()}')"

    def __repr__(self):
        return f"Marker('{self.nm()}')"


class Attributes(Base):
    """[script.markdown.attributes]"""

    # fmt: off
    excluded: List[str] = Field(
        default_factory=list, alias='exclude'
    )
    from_namespace: Dict[str, str] = Field(
        default_factory=dict, alias='from-namespace'
    )
    groups: Dict = Field(
        default_factory=dict, alias="groups"
    )
    order: List[str] = Field(
        default_factory=list, alias="attribute-order"
    )
    reserved: Dict[str, Reserved] = Field(
        default_factory=dict, alias="reserved"
    )
    markers: Dict[str, Marker] = Field(
        default_factory=dict, alias="attribute-markers"
    )
    # fmt: on

    def __init__(self, **data):

        super().__init__(**data)

        self.order = data.get("order", dict()).get("attribute-order", list())

        for attr_nm, defaults in data["reserved"].items():
            args = {"attribute-name": attr_nm}
            self.reserved[attr_nm] = Reserved(**{**args, **defaults})

        # markers = data.get("markers")
        for marker, marker_attrs in data["markers"].items():
            self.markers[marker] = Marker(**marker_attrs).set_name(name=marker)

    def exclude(self, item: str):
        """Adds an item (argument name) to list of exclusions."""
        if item not in self.excluded:
            self.excluded.append(item)
        return self

    def get_marker(self, name: str):
        """Fetches a template marker from :attr:`markers`."""
        return self.markers.get(f"__{name}__")

    def merge_markers(self, parsed_markers: Dict[int, Dict]) -> Dict[int, Marker]:
        """Merges markers parsed from script with template markers in snowmobile.toml.

        Does the following:
            *   Consumes all parsed attributes from markers found in a script
            *   Tries to pull a pre-configured marker based on its time and
                updates its pre-configured values with those pr in the
                script if so
            *   If it doesn't find a pre-configured marker based on the marker
                name, it will instantiate a new marker instance from the
                arguments pr in the script.

        Args:
            parsed_markers (Dict[int, Dict]):
                All parsed raw marker arguments from a script by index position.

        Returns (Dict[int, Marker]):
            Instantiated markers for all attributes, merging those with
            matching names to pre-configured markers in snowmobile.toml.

        """
        total = {}
        for i, ms in parsed_markers.items():
            m = self.get_marker(name=ms["marker-name"])
            m = m.update(attrs=ms) if isinstance(m, Marker) else Marker(**ms)
            total[i] = m
        return total

    def get_position(self, attr: str) -> int:
        """Returns the position for an attribute based on snowmobile-ext.toml.

        Will return 0 if not included in order-by configuration.

        """
        attr = attr.lower()
        ordered = [w.lower() for w in self.order]
        if attr not in ordered:
            return int()
        for i, ordered_attr in enumerate(ordered, start=1):
            if attr == ordered_attr:
                return i

    def included(self, attrs: Dict) -> Dict:
        """Checks if an attribute has been marked for exclusion from render.
        """
        return {k: v for k, v in attrs.items() if k not in self.excluded}

    def group_parsed_attrs(self, parsed: Dict) -> Dict:
        """Nests attributes into dictionaries that are configured as groups.
        """
        grouped = {}
        attrs = self.included(parsed)
        for parent, child_attrs in self.groups.items():
            children = {
                attr: attrs.pop(attr) for attr in child_attrs if attrs.get(attr)
            }
            if children:
                grouped[parent] = children
        return {**grouped, **attrs}

    def _add_reserved(self, nm: str, attrs: Dict, is_marker: bool = False):
        """Adds a single reserved attr's configuration to wrap.

        Args:
            nm: Attribute name.
            attrs: Full dictionary of attributes.
            is_marker: Dictionary of attributes is for a marker.

        Returns: Revised wrap.

        """
        reserved_attr = self.reserved[nm]
        reserved_nm = reserved_attr.attr_nm
        if is_marker:
            return attrs
        if not attrs.get(reserved_nm) and reserved_attr.include_by_default:
            attrs[reserved_attr.default_val] = reserved_attr.include_by_default
        elif reserved_attr.include_by_default:
            attrs[reserved_attr.default_val] = attrs[reserved_nm]
        return attrs

    def add_reserved_attrs(self, attrs: Dict, is_marker: bool = False):
        """Batch modifies all reserved attributes to their configuration."""
        attrs = {k: v for k, v in attrs.items()}  # work on a copy of wrap
        for nm in self.reserved:
            attrs = self._add_reserved(nm=nm, attrs=attrs, is_marker=is_marker)
        return attrs


class Core(Base):
    """[script.patterns.core]"""

    # fmt: off
    to_open: str = Field(
        default_factory=str, alias='open-tag'
    )
    to_close: str = Field(
        default_factory=str, alias='close-tag'
    )
    delimiter: str = Field(
        default_factory=str, alias="description-delimiter"
    )
    prefix: str = Field(
        default_factory=str, alias="description-index-prefix"
    )
    # fmt: on


class Markup(Base):
    """[script.markup]"""

    # fmt: off
    hx_marker: str = Field(
        default_factory=str, alias='default-marker-header'
    )
    hx_statement: str = Field(
        default_factory=str, alias='default-statement-header'
    )
    bullet_char: str = Field(
        default_factory=str, alias='default-bullet-character'
    )
    attr_nm_wrap_char: str = Field(
        default_factory=str, alias="wrap-attribute-names-with"
    )
    attr_value_wrap_char: str = Field(
        default_factory=str, alias="wrap-attribute-values-with"
    )
    incl_index_in_sh: bool = Field(
        default_factory=bool, alias='include-statement-index-in-header'
    )
    # result_limit: int = Field(
    #     default_factory=int, alias="limit-query-results-to"
    # )
    attrs: Attributes = Field(
        default_factory=Attributes, alias="attributes"
    )
    # fmt: on

    def pref_header(
        self, is_marker: bool = False, from_wc: Optional[str] = False
    ) -> str:
        """Creates header prefix based on specifications."""
        h_level = self.hx_marker if is_marker else self.hx_statement
        if not from_wc:
            return int(h_level.strip()[1:]) * "#"
        wc_only = "".join(c for c in from_wc if c == "*")
        return len(wc_only) * "#"


class Pattern(Base):
    """[script.patterns]"""

    # fmt: off
    core: Core = Field(
        default_factory=Core, alias="core"
    )
    wildcards: Wildcard = Field(
        default_factory=Wildcard, alias="markup"
    )
    # fmt: on


class Tolerance(Base):
    """[script.qa.default-tolerance]"""

    # fmt: off
    relative: float = Field(
        default_factory=float, alias="relative"
    )
    absolute: float = Field(
        default_factory=float, alias="absolute"
    )
    only_matching_rows: bool = Field(
        default_factory=bool, alias="only-compare-matching-rows"
    )
    # fmt: on


class QA(Base):
    """[script.qa]"""

    # fmt: off
    partition_on: str = Field(
        default_factory=str, alias="partition-on"
    )
    ignore_patterns: List = Field(
        default_factory=list, alias="ignore-patterns"
    )
    compare_patterns: List = Field(
        default_factory=list, alias="compare-patterns"
    )
    end_index_at: str = Field(
        default_factory=str, alias="end-index-at"
    )
    tolerance: Tolerance = Field(
        default_factory=Tolerance, alias="default-tolerance"
    )
    # fmt: on


class Type(Base):
    """snowmobile-ext.toml: [wrap-to-type-xref]"""

    # fmt: off
    as_str: List = Field(
        default_factory=list, alias="string"
    )
    as_list: List = Field(
        default_factory=list, alias="list"
    )
    as_float: List = Field(
        default_factory=list, alias="float"
    )
    as_bool: List = Field(
        default_factory=list, alias="bool"
    )
    # fmt: on


class Script(Base):
    """[script]"""

    # fmt: off
    patterns: Pattern = Field(
        default_factory=Pattern, alias="patterns"
    )
    markup: Markup = Field(
        default_factory=Markup, alias="markup"
    )
    qa: QA = Field(
        default_factory=QA, alias="qa"
    )
    types: Type = Field(
        default_factory=Type, alias="wrap-to-type-xref"
    )
    export_dir_nm: str = Field(
        default_factory=str, alias="export-dir-name"
    )
    result_limit: int = Field(
        default_factory=int, alias="result-limit"
    )
    # fmt: on
    
    def tag(self) -> Tuple[str, str]:
        """Open/close pattern for statement tags."""
        return (
            self.patterns.core.to_open,
            self.patterns.core.to_close,
        )

    @staticmethod
    def power_strip(val_to_strip: str, chars_to_strip: Iterable[str]) -> str:
        """Exhaustively strips a string of specified leading/trailing characters."""
        while any(val_to_strip.strip(c) != val_to_strip for c in chars_to_strip):
            for c in chars_to_strip:
                val_to_strip = val_to_strip.strip(c)
        return val_to_strip

    def arg_to_string(self, arg_as_str: str) -> str:
        """Strips an argument as a string down to its elemental form."""
        return self.power_strip(val_to_strip=arg_as_str, chars_to_strip=["'", '"', " "])

    def arg_to_list(self, arg_as_str: str) -> List[str]:
        """Converts a list as a string into a list."""
        return [
            self.power_strip(v, chars_to_strip=['"', "'", "[", "]", " "])
            for v in arg_as_str.split(",")
        ]

    def arg_to_float(self, arg_as_str: str) -> float:
        """Strips a string down to its elemental form and converts to a float."""
        return float(self.arg_to_string(arg_as_str))

    def arg_to_bool(self, arg_as_str: str) -> bool:
        """Converts a boolean in string-form into a boolean value."""
        str_replacements = {"true": True, "false": False}
        return str_replacements[self.arg_to_string(arg_as_str).lower()]

    def parse_arg(self, arg_key: str, arg_value: str):
        """Parses an argument into its target data type based on its `arg_key`
        and the ``script.name-to-type-xref`` defined in **snowmobile.toml**."""
        arg_key, _, _ = arg_key.partition("*")
        if arg_key in self.types.as_list:
            return self.arg_to_list(arg_as_str=arg_value)
        elif arg_key in self.types.as_float:
            return self.arg_to_float(arg_as_str=arg_value)
        elif arg_key in self.types.as_bool:
            return self.arg_to_bool(arg_as_str=arg_value)
        else:
            return self.arg_to_string(arg_as_str=arg_value)

    @staticmethod
    def split_args(args_str: str) -> List[str]:
        """Returns a list of arguments based on splitting string on double
        underscores and stripping results."""
        matches = [
            s.strip() for s in re.split(r"^__", args_str, flags=re.MULTILINE)
        ]
        return [s for s in matches if not s.isspace() and s]

    def parse_split_arguments(self, splitter: List[str]) -> Dict:
        """Returns a dictionary of argument-index to argument keys and values."""
        args_parsed = {}
        for i, s in enumerate(splitter, start=1):
            keyword, _, arg = s.partition(":")
            if keyword and arg:
                args_parsed[keyword.strip()] = self.parse_arg(
                    arg_key=keyword, arg_value=arg
                )
        return args_parsed

    def parse_str(
        self, block: str, strip_blanks: bool = False, strip_trailing: bool = False
    ) -> Dict:
        """Parses a string of statement tags/arguments into a valid dictionary.

        Args:
            block (str):
                Raw string of all text found between a given open/close wrap.
            strip_blanks (bool):
                Strip blank lines from string; defaults to `False`.
            strip_trailing (bool):
                Strip trailing whitespace from the string; defaults to `False`.

        Returns (dict):
            Dictionary of arguments.

        """
        stripped = p.strip(block, blanks=strip_blanks, trailing=strip_trailing)
        splitter = self.split_args(args_str=stripped)
        return self.parse_split_arguments(splitter=splitter)

    def wrap(self, tag: str) -> str:
        """Wraps a raw string of sql in open/closing patterns."""
        _open, _close = self.tag()
        return f"{_open}{tag}{_close}"
    
    @staticmethod
    def attr_construct(attr_nm: str, attr_value: str) -> str:
        """Returns a parsable attribute from an attribute name and value."""
        return f"__{attr_nm}: {attr_value}"
    
    def tag_from_attrs(
        self, attrs: Dict, nm: Optional[str] = None, wrap: bool = False, **kwargs
    ) -> str:
        """Construct a parsable tag out of a dictionary of attributes."""
        if not attrs.get('name', nm) and not kwargs.get('unsafe'):
            raise ValueError(
                "'name' must be part of 'attrs' or provided as a method argument."
            )
        attrs = attrs.copy()
        
        if nm:
            attrs['name'] = nm
            
        _attrs = [self.attr_construct('name', attrs.pop('name'))]
        ordered_attributes = self.markup.attrs.order
        
        for attr in ordered_attributes:
            if attrs.get(attr):
                _attrs.append(self.attr_construct(attr, attrs[attr]))
        for attr_nm, attr_value in attrs.items():
            if attr_nm != 'name' and attr_nm not in ordered_attributes:
                _attrs.append(self.attr_construct(attr_nm, attr_value))
                
        _tag = '\n'.join(_attrs)
        tag = f"\n{_tag}\n"
        return self.wrap(tag) if wrap else tag

    def as_parsable(
        self,
        raw: str,
        is_multiline: Optional[bool] = None,
        is_marker: Optional[bool] = None,
        lines: Optional[int] = None,
    ) -> str:
        """Returns a raw string wrapped in open/close tags.

        Used for taking a raw string of marker or statement attributes and
        wrapping it in open/close tags before exporting, making the script
        being exported re-parsable by `snowmobile`.

        """
        raw = raw.strip("\n")
        _open = self.patterns.core.to_open
        _close = self.patterns.core.to_close
        if is_marker:
            lines = (0 if lines == -1 else lines or 1) * "\n"
            return f"{_open}\n{raw}\n{_close}{lines}"
        elif is_multiline:
            return f"{_open}\n{raw}\n{_close}"
        return f"{_open}{raw}{_close}"

    def find_spans(self, sql: str) -> Dict[int, Tuple[int, int]]:
        """Finds indices of script tags given a sql script as a string and an
        open and close pattern of the tags."""

        to_open, to_close = self.patterns.core.to_open, self.patterns.core.to_close
        _open, _close = re.compile(re.escape(to_open)), re.compile(re.escape(to_close))

        open_spans = {
            i: m.span()[1] for i, m in enumerate(re.finditer(_open, sql), start=1)
        }
        close_spans = {
            i: m.span()[0] for i, m in enumerate(re.finditer(_close, sql), start=1)
        }
        try:
            if len(open_spans) != len(close_spans):
                raise AssertionError(
                    f"parsing.find_tags() error.\n"
                    f"Found different number of open-tags to closing-tags; please check "
                    f"script to ensure each open-wrap matching '{to_open}' has an "
                    f"associated closing-_tag matching '{to_close}'."
                )
            return {i: (open_spans[i], close_spans[i]) for i in close_spans}
        except AssertionError as e:
            raise e

    def find_tags(self, sql: str) -> Dict[int, str]:
        """Finds indices of script tags given a sql script as a string and an
        open and close pattern of the tags."""
        try:
            bounded_arg_spans_by_idx = self.find_spans(sql=sql)
            return {
                i: sql[span[0]: span[1]]
                for i, span in bounded_arg_spans_by_idx.items()
            }
        except AssertionError as e:
            raise e

    def find_block(self, sql: str, marker: str) -> str:
        """Finds a block of arguments based on a marker.

        Markers expected by default are the __script__ and __appendix__ markers.

        """
        tags = [t for t in self.find_tags(sql=sql).values() if marker in t]
        assert len(tags) <= 1, (
            f"Found more than one wrap within script containing '{marker}'; "
            f"expected exactly one."
        )
        return tags[0] if tags else str()

    def has_tag(self, s: sqlparse.sql.Statement) -> bool:
        """Checks if a given statement has a wrap that directly precedes the sql."""
        s_tot: str = s.token_first(skip_cm=False).value
        spans_by_index = self.find_spans(sql=s_tot)
        if not spans_by_index:
            return False

        _, last_span_close_idx = spans_by_index[max(spans_by_index)]
        s_remainder = s_tot[last_span_close_idx:]
        if not s_remainder.startswith(self.patterns.core.to_close):
            raise Exception(
                f"Last span identified in the below statement did not end with"
                f" the expected close-pattern of {self.patterns.core.to_close}"
                f"; see sql below.\n\n{s_tot}"
            )

        return len(s_remainder.split("\n")) == 2

    @staticmethod
    def is_marker(raw: str):
        """Checks if a raw string of arguments has a marker on the first line."""
        if not raw:
            return bool(raw)
        first_line = [v for v in raw.split("\n") if v][0]
        return bool(re.findall("__.*__", first_line))

    @staticmethod
    def is_valid_sql(s: sqlparse.sql.Statement) -> bool:
        """Verifies that a given :class:`sqlparse.sql.Statement` contains valid sql."""
        has_sql = bool(s.token_first(skip_cm=True, skip_ws=True))
        return has_sql and not s.value.isspace()

    @staticmethod
    def strip_comments(s: sqlparse.sql.Statement) -> str:
        """Isolates just the sql within a :class:`sqlparse.sql.Statement` object."""
        s_sql = s.token_first(skip_cm=True)
        if not s_sql:
            return str()
        first_keyword_idx = s.token_index(s_sql)
        sql = "".join(c.value for c in s[first_keyword_idx:])
        return sql.strip().strip(";")

    def split_sub_blocks(self, s: sqlparse.sql.Statement) -> Tuple[List, str]:
        """Breaks apart blocks of arguments within a :class:`sqlparse.sql.Statement`.

        Note:
            *   :meth:`sqlparse.parsestream()` returns a list of
                :class:`sqlparse.sql.Statement` objects, each of which includes
                all text (comments) between the last character of the prior
                statement and the first character of the current one.
            *   :meth:`split_sub_blocks()` traverses that space and identifies
                all spans of text wrapped in `open` (``/*-``) and `close`
                (``-*/``) tags, storing their index positions relative to the
                other statements & markers.
            *   These are stored as :class:`snowmobile.core.Script` attributes
                as statements are parsed and so that they can be exported in
                the appropriate order to a markdown file.

        Args:
            s (sqlparse.sql.Statement):
                :class:`sqlparse.sql.Statement` object.

        Returns (Tuple[List, str]):
            A tuple containing:
                1.  A list of __marker__ blocks if they exist; empty list otherwise
                2.  The last wrap/block before the start of the actual SQL (e.g. the
                    wrap/block that is associated with the statement passed to ``s``.

        """
        blocks = list(self.find_tags(sql=s.value).values())
        markers = [b for b in blocks if self.is_marker(raw=b)]
        is_tagged = self.has_tag(s=s)
        tag = blocks[-1] if is_tagged else str()
        return markers, tag

    def name_from_marker(self, raw: str) -> str:
        """Extracts a marker name (e.g. 'script' from within __script__)."""
        splitter = self.split_args(args_str=raw)
        if not splitter:
            raise Exception(
                f"snowmobile parsing error: parsing.name_from_marker() called on"
                f"an empty string."
            )
        valid = [v for v in splitter[0].partition('__') if v]
        return ''.join(valid[:-1]) if len(valid) > 1 else valid[0]
        # return splitter[0]

    def parse_name(
        self, raw: str, offset: Optional[int] = None, silence: bool = False
    ) -> str:
        """Parses name from a raw set of arguments if not given an explicit wrap."""
        by_line = raw.strip("\n").split("\n")
        offset = offset or 0
        try:
            if by_line[offset].startswith("__"):
                e = errors.InvalidTagsError(
                    msg=f"""
    invalid statement tags pr.
    multi-line statement tags without an explicit `__name` attribute must include
    a name not beginning with '__' on the first line within the open & closing wrap;
    first line found is:\n```\n{raw}\n```.
    """
                )
                if silence:
                    return str()
                raise e
        except IndexError as e:
            e2 = errors.InvalidTagsError(
                msg=f"Error encountered while trying to parse the following:\n\t{raw.strip()}"
            )
            raise e2 from e
        return self.power_strip(by_line[offset], ['"', "'", " "])

    @staticmethod
    def add_name(nm_title: str, nm_marker: str, attrs: dict, overwrite: bool = False):
        """Adds a name to a set of parsed marker attributes.

        Accepts a name and a dict of parsed attributes from a marker and:
            1.  Checks to see if there's an explicit 'name' declared within the
                attributes
            2.  If not explicitely declared **or** explicitely declared and
                `overwrite=False`, it will add the `nm` value to the attributes as 'name'.
            3.  It will also add the 'nm' value to the attributes as 'marker-name'
                to be used by the :class:`Marker` when cross-referencing the
                __name__ with template markers in ``snowmobile.toml``.

        Args:
            nm_title (str):
                The name of the marker as either:
                    1.  Returned value from :meth:`name_from_marker()`
                    2.  Returned value from :meth:`parse_name()`
                    3.  None if neither is pr
            nm_marker (str):
                The string value wrapped in ``__`` on the first line of the
                argument block.
            attrs (dict):
                A dictionary of parsed attributes as returned from :func:`parse_str()`.
            overwrite (bool):
                Indicator of whether or not to overwrite a 'name' attribute declared
                within the .sql script.

        """
        attrs["marker-name"] = nm_marker
        if nm_title:
            attrs["name"] = nm_title
        if not nm_title or overwrite:
            attrs["name"] = nm_marker
        return attrs

    def parse_marker(self, attrs_raw: str) -> Dict:
        """Parses a raw string of __marker__ text between an open and a close pattern."""
        parsed = self.parse_str(attrs_raw)
        nm_title = parsed.get("name") or self.parse_name(
            raw=attrs_raw, offset=1, silence=True
        )
        marker_nm = self.name_from_marker(attrs_raw)
        self.add_name(
            nm_title=nm_title, nm_marker=marker_nm, attrs=parsed, overwrite=False
        )
        parsed["raw-text"] = attrs_raw
        return parsed

    @staticmethod
    def ensure_sqlparse(
        sql: Union[sqlparse.sql.Statement, str]
    ) -> sqlparse.sql.Statement:
        """Returns a :class:`sqlparse.sql.Statement` from ``sql``.

        Will return ``sql`` with no modification if it's already a
        :mod:`sqlparse.sql` object.

        Needed to accommodate dynamic addition of statements as strings to
        an existing :class:`~snowmobile.Script` object from
        from raw strings as opposed to a :class:`sqlparse.sql.Statement`
        objects as is done when reading a sql file.

        Args:
            sql (Union[sqlparse.sql.Statement, str]):
                Either a string of sql or an already parsed
                sqlparse.sql.Statement object.

        Returns (sqlparse.sql.Statement):
            A parsed sql statement.

        """
        if isinstance(sql, sqlparse.sql.Statement):
            return sql

        parsed = [
            s
            for s in sqlparse.parsestream(stream=sql.strip().strip(";"))
            if not s.value.isspace()
        ]
        assert (
            parsed
        ), f'sqlparse.parsestream("""\n{sql}"""\n) returned an empty statement.'
        return parsed[0]

    def sql_tokens(self, sql: str) -> List[sqlparse.sql.Token]:
        """Unpacks nested tokens from a sqlparse.sql.Statement.

        Args:
            sql (str): A raw sql from a statement.

        Returns:
            A list of tokens.

        """
        s = self.ensure_sqlparse(sql=sql)
        id_tokens = [
            [
                [t] if (
                    not hasattr(t, 'tokens')
                    or isinstance(t, sqlparse.sql.Identifier)
                )
                else [t2 for t2 in t.tokens]
                for t in [t3 for t3 in s.tokens]
            ]
        ]
        unpacked_intl = [t_sub for t in id_tokens for t_sub in t]
        return [
            t
            for t in itertools.chain.from_iterable(unpacked_intl)
            if not self._is_whitespace(t)
        ]

    def id_from_tokens(self, sql: str) -> str:
        """Identifies the last identifier in a piece of raw sql.

        Identifies `obj` being operated on.

        Args:
            sql (str): A raw piece of sql from a statement.

        Returns:
            A string if the last identifier found or an empty string otherwise.

        """
        ids = [
            t for t in self.sql_tokens(sql)
            if isinstance(t, sqlparse.sql.Identifier)
            or self._is_name(t)
        ]
        if 'clone' not in sql.lower().split('\n')[0]:
            return ids[-1].value.split(' ')[0] if ids else str()
        vals = " ".join(i.value for i in ids).split(' ')
        return " ".join(vals[-3:]) if len(vals) >= 3 else str()

    @staticmethod
    def _is_name(t: sqlparse.sql.Token) -> bool:
        """Is name token."""
        return 'name' in str(t.ttype).lower()

    @staticmethod
    def _is_whitespace(t: sqlparse.sql.Token) -> bool:
        """Is name token."""
        return 'whitespace' == str(t.ttype).lower().split('.')[-1]



"""
Instantiate `script` from 'overview.sql' and inspect high-level contents.
../docs/snippets/script/overview-base-parsing.py
"""

# Setup -----------------------------------------------------------------------
from pathlib import Path
paths = {
    p.name: p
    for p in Path.cwd().glob('**/*.sql')
    if '.snowmobile' not in p.parents
}
path = paths['overview-base-sn.sql']

import snowmobile


# Example ---------------------------------------------------------------------

# -- Block 1 --
script = snowmobile.Script(path=path)
script.dtl()
"""
>>>
overview.sql
============
1: Statement('create table~s1')
2: Statement('insert into~s2')
3: Statement('select data~s3')
4: Statement('select all~sample_table')
5: Statement('create transient table~s5')
6: Statement('insert into~s6')
7: Statement('drop table~s7')
"""
#
# script.run(on_failure='c')
# script(4).desc
# script.run((1, -3))
#
# script.sn.query(script(10).sql)
# script(9).attrs_parsed
# script.dtl()
#
# script.run(8, on_failure='c')
# script(8).results.head()
#
# script.sn.drop('any_other_table')
# script.sn.drop('sample_table')
#
# script.dtl()
# script(3).run().results.head()
# script(8).run().results.head()
#
# with script.filter(incl_desc='DDL') as s:
#     s.dtl()

# with script.filter(
#     excl_anchor=['qa.*', 'sample\b.*'],
#     excl_desc=['DDL'],
#     excl_kw='truncate',
# ) as s:
#     s.dtl()

# -- Block 2 --
# Store a few st, accessed by index position
s_first, s_last = script(1), script(-1)

# first sql keyword
print(s_first.kw)  #> create
print(s_last.kw)   #> drop

# position within `script`
print(s_first.index)  #> 1
print(s_last.index)   #> 7

# -- Block 3 --
script.run(1)    # .run() from `script`
script(1).run()  # .run() from `statement`

# -- Block 4 --
# `script` details as read from 'overview.sql'
print(script.depth)   #> 7
print(script(1).nm)   #> create table~s1
print(script(-1).nm)  #> drop table~s7

with script.filter(excl_kw=['select', 'drop']) as s:
    print(s.depth)   #> 4
    print(s(1).nm)   #> create table~s1
    print(s(-1).nm)  #> insert into~s4
    s.dtl()
"""
>>>
overview.sql
============
1: Statement('create table~s1')
2: Statement('insert into~s2')
3: Statement('create transient table~s3')
4: Statement('insert into~s4')
"""



"""
Tests for snowmobile.statement.Scope.
"""
import pytest


@pytest.fixture()
def sample_scope_object(sn_delayed):
    """An example statement object for testing."""
    from snowmobile.core.scope import Scope

    return Scope(arg="kw", base="select")


@pytest.mark.scope
def test_scope_defaults(sample_scope_object):
    """Tests default values of a scope object."""
    assert sample_scope_object.component == "kw"
    assert sample_scope_object.base == "select"

    assert sample_scope_object.incl_arg == "incl_kw"
    assert sample_scope_object.excl_arg == "excl_kw"

    assert not sample_scope_object.is_excluded
    assert not sample_scope_object.is_included
    assert not sample_scope_object

    assert sample_scope_object.fallback_to == {"incl_kw": ["select"], "excl_kw": []}


@pytest.mark.scope
def test_scope_dunder_methods(sample_scope_object):
    """Tests dunder methods for a scope object."""
    assert repr(sample_scope_object) == "Scope(arg='kw', base='select')"
    assert str(sample_scope_object) == "Scope(arg='kw', base='select', included=False)"
    assert not bool(sample_scope_object)



"""Tests for snowmobile.utils.parsing."""
import pytest


def setup_schema_parsing_test():
    """Sets up values and IDs for test of parsing.p()"""
    # (input_to_test, tuple[expected_schema, expected_name])
    input_to_expected_output = [
        ("object", ("", "object")),
        (" object ", ("", "object")),
        ("schema.object", ("schema", "object")),
        ("__schema", ("schema", "")),
        (None, ("", "")),
    ]
    ids = [
        f"in='{i[0]}', out=('{i[1][0]}', '{i[1][1]}')" for i in input_to_expected_output
    ]
    return ids, input_to_expected_output


ids, test_cases = setup_schema_parsing_test()


@pytest.mark.parsing
@pytest.mark.parametrize("names", test_cases, ids=ids)
def test_parsing_schema_and_name_from_generic_nm(names):
    """Test parsing.p() produces expected as_df given a nm (str)."""
    from snowmobile.core.utils.parsing import p

    nm, output_expected = names
    output_under_test = p(nm=nm)
    assert output_under_test == output_expected



"""
Decomposes a statement into parts and generates a name given its content
and position within the script.
"""
from __future__ import annotations

from typing import Optional, Set

from . import Generic, Configuration, Scope
from .cfg import Pattern
from .tag import Attrs


class Name(Generic):
    """Handles the decomposition/parsing of statement name.

    Should never be instantiated directly by the user-facing API but its
    attributes are likely to be accessed often as part of :class:`Statement`
    and derived classes.

    Attributes:
        cfg (snowmobile.Configuration):
            :class:`snowmobile.Configuration` object; represents fully parsed
            **snowmobile.toml** file.
        patt (snowmobile.Schema.Pattern):
            :class:`snowmobile.Schema.Pattern` object; represents
            ``script.patterns`` section of **snowmobile.toml**.
        _nm_pr (str):
            Provided wrap name for a given :class:`Statement`; can be empty.
        index (int):
            Statement index position within :class:`Script`; can be empty.
        is_included (bool):
            Indicator of whether or not the combination of all scopes for this
            statement wrap is included within a given context.
        incl_idx_in_desc (bool):
            Indicator of whether or not to include the statement index in the
            `description` component of the wrap; defaults to `True` so that all
            ge statement tags are guaranteed to be unique for a given
            script.
                *   Mainly included for testing purposes where setting to
                    `False` enables comparing ge to pr statement
                    tags without having to change the index position of the
                    hard-coded/pr statement wrap when adding/removing tests.
        first_line_remainder (str):
            The remainder of the first line once excluding the
            :attr:`first_keyword` and stripping repeating whitespace.
        scopes (set[Scope]):
            Combination of all scopes for a given wrap; this is essentially the
            all possible combinations of including/excluding any of the `kw`,
            `nm`, `obj`, `desc`, and `anchor` for a given instance of :class:`Name`.

    """

    def __init__(
        self,
        configuration: Configuration,
        nm_pr: Optional[str] = None,
        sql: Optional[str] = None,
        index: Optional[int] = None,
        # attrs: Optional[Attrs] = None,
    ):
        super().__init__()

        # configuration
        # -------------
        self._code = sql
        self.cfg: Configuration = configuration
        self.patt: Pattern = self.cfg.script.patterns
        self.index = index or int()
        self.is_included: bool = True
        self.incl_idx_in_desc: bool = True

        # '_pr' placeholders
        # ------------------
        self._nm_pr = nm_pr or str()
        # self._attrs = attrs or dict()
        self._anchor_pr = str()
        self._kw_pr = str()
        self._obj_pr = str()
        self._desc_pr = str()

        # '_pr' parsing
        # -------------
        self.part_desc = tuple(
            v for v in self._nm_pr.partition(self.patt.core.delimiter) if v
        )
        if len(self.part_desc) == 3:  # ('create table', '~', 'sample_records')
            anchor_vf = [  # ensure no extra space between words
                self.cfg.script.power_strip(v, " ")
                for v in self.part_desc[0].split(" ")
                if self.cfg.script.power_strip(v, " ")
            ]
            self._anchor_pr = " ".join(anchor_vf)  # 'create table'
            self._kw_pr = anchor_vf[0]  # 'create'
            self._obj_pr = (  # 'table'
                " ".join(anchor_vf[1:]) if len(anchor_vf) >= 2 else str()
            )
            self._desc_pr = " ".join(  # 'sample records'
                self.cfg.script.power_strip(v, " ")
                for v in self.part_desc[-1].split(" ")
                if self.cfg.script.power_strip(v, " ")
            )

        # sql parsing
        # -----------
        stripped_code = self.cfg.script.power_strip(
            val_to_strip=sql, chars_to_strip="\n "  # trailing lines and whitespace
        )
        self.first_line = self.cfg.script.power_strip(
            val_to_strip=stripped_code.split("\n")[0].lower(),
            chars_to_strip="\n ",  # same for first line only
        )
        self.words_in_first_line = [
            self.cfg.script.arg_to_string(v)
            for v in self.first_line.split(" ")
            if self.cfg.script.arg_to_string(v)
        ]
        self.first_line_remainder = " ".join(
            self.words_in_first_line[1:]
            if len(self.words_in_first_line) >= 2
            else str()
        )
        self.matched_terms = self.cfg.sql.objects_within(
            self.first_line  # finds keywords by full term within first line
        )

        # then create scope
        # -----------------
        self.scopes: Set[Scope] = {
            Scope(**kwargs) for kwargs in self.cfg.scopes_from_tag(t=self)
        }

    def _pr_parse(self):
        """Parses provided name into core components."""
        self.part_desc = tuple(
            v for v in self._nm_pr.partition(self.patt.core.delimiter) if v
        )
        if len(self.part_desc) == 3:  # ('create table', '~', 'sample_records')
            anchor_vf = [  # ensure no extra space between words
                self.cfg.script.power_strip(v, " ")
                for v in self.part_desc[0].split(" ")
                if self.cfg.script.power_strip(v, " ")
            ]
            self._anchor_pr = " ".join(anchor_vf)  # 'create table'
            self._kw_pr = anchor_vf[0]  # 'create'
            self._obj_pr = (  # 'table'
                " ".join(anchor_vf[1:]) if len(anchor_vf) >= 2 else str()
            )
            self._desc_pr = " ".join(  # 'sample records'
                self.cfg.script.power_strip(v, " ")
                for v in self.part_desc[-1].split(" ")
                if self.cfg.script.power_strip(v, " ")
            )

    def scope(self, **kwargs) -> bool:
        """Evaluates all component's of a wrap's scope against a set of filter args.

            **kwargs:
                Keyword arguments passed to :class:`Script.filter()` (e.g.
                `incl_kw`, `excl_kw`, ..)

        Returns (bool):
            Value indicating whether or not the statement should be included
            based on the outcome of the evaluation of all of its components.

        """
        self.is_included = all(s.eval(**kwargs) for s in self.scopes)
        return self.is_included

    @property
    def _kw_ge(self):
        """Generated `keyword` for statement."""
        return (
            self.cfg.sql.kw_exceptions.get(
                self.words_in_first_line[0], self.words_in_first_line[0]
            )
        )

    @property
    def _obj_ge_base(self):
        """Base for ge object."""
        non_overlapping = {
            i: t
            for i, t in self.matched_terms.items()
            if t != self._kw_ge and self._kw_ge not in ['set', 'unset']
        }
        return (
            non_overlapping[min(non_overlapping)] if non_overlapping else str()
        )

    @property
    def _obj_ge(self):
        """Generated `object` for statement."""
        return (
            self._obj_ge_base or
            (
                self._anchor_ge.split(' ')[-1]
                if len(self._anchor_ge.split(' ')) > 1 else str()
            )
        )

    @property
    def _desc_ge(self) -> str:
        """Generated `description` for statement."""
        idx_prefix = self.cfg.script.patterns.core.prefix
        if self.cfg.sql.desc_is_simple:
            return (
                f"{idx_prefix}{self.index}" if self.incl_idx_in_desc
                else 'statement'
            )
        _desc = self.cfg.script.id_from_tokens(sql=self.first_line)
        desc = f"{_desc}" if _desc else str()
        return (
            f"{desc}: {idx_prefix}{self.index}" if self.incl_idx_in_desc
            else {desc}
        )

    @property
    def _anchor_ge(self):
        """Generated `anchor` for statement."""
        generalized_anchor = self.cfg.sql.generic_anchors.get(self._kw_ge)
        if generalized_anchor and self._obj_ge_base == str():
            return generalized_anchor
        s = " " if self._kw_ge and self._obj_ge_base else ""
        return f"{self._kw_ge}{s}{self._obj_ge_base}"

    @property
    def _nm_ge(self):
        """Generated `name`; all delimiters included."""
        return f"{self._anchor_ge}{self.patt.core.delimiter}{self._desc_ge}"

    def _nm(self, og: bool) -> str:
        """Constructing 'nm' from raw attributes if og=False."""
        delim = self.cfg.script.patterns.core.delimiter
        return (
            self._nm_pr
            if original             else
            f"{self._anchor(og)}{delim}{self.desc()}"
        )
        
    def nm(
        self, ge: bool = False, pr: bool = False, og: bool = False
    ) -> str:
        """The final statement's **name** that is used by the API.

        This will be the full wrap name if a statement wrap exists and a
        parsed/ge wrap name otherwise.

        """
        if ge:
            return self._nm_ge
        if pr:
            return self._nm(og=og)
        if self.cfg.sql.pr_over_ge:
            return self._nm(og=og) or self._nm_ge
        return self._nm_ge

    def kw(self, ge: bool = False, pr: bool = False):
        """The final statement's **keyword** that is used by the API.

        This will be the pr keyword if a statement wrap exists and a
        parsed/ge keyword otherwise.

        """
        if ge:
            return self._kw_ge
        if pr:
            return self._kw_pr
        if self.cfg.sql.pr_over_ge:
            return self._kw_pr or self._kw_ge
        return self._kw_ge

    def obj(self, ge: bool = False, pr: bool = False):
        """The final statement's **object** that is used by the API.

        This will be the object within a wrap if a statement wrap exists and
        follows the correct structure and a parsed/ge object otherwise.

        """
        if ge:
            return self._obj_ge
        if pr:
            return self._obj_pr
        if self.cfg.sql.pr_over_ge:
            return self._obj_pr or self._obj_ge
        return self._obj_ge

    def desc(self, ge: bool = False, pr: bool = False):
        """The final statement's **description** that is used by the API.

        This will be the description within a wrap if a statement wrap exists
        and follows the correct structure and a parsed/ge description
        otherwise.

        """
        if ge:
            return self._desc_ge
        if pr:
            return self._desc_pr
        if self.cfg.sql.pr_over_ge:
            return self._desc_pr or self._desc_ge
        return self._desc_ge

    def _anchor(self, og: bool) -> str:
        """Constructing 'anchor' from raw attributes if og=False."""
        return (
            self._nm_pr
            if original             else
            f"{self.kw()} {self.obj()}"
        )

    def anchor(self, ge: bool = False, pr: bool = False):
        """The final statement's **anchor** that is used by the API.

        This will be the anchor within a wrap if a statement wrap exists
        and follows the correct structure and a parsed/ge wrap name
        otherwise.

        """
        if ge:
            return self._anchor_ge
        if pr:
            return self._anchor_pr
        if self.cfg.sql.pr_over_ge:
            return self._anchor_pr or self._anchor_ge
        return self._anchor_ge

    def set(self, key, value) -> Name:
        """Custom attribute setting."""
        attrs = vars(self)
        if f"_{key}_pr" in attrs:
            key = f"_{key}_pr"
        vars(self)[key] = value
        return self

    def __bool__(self) -> bool:
        return self.is_included

    def __str__(self) -> str:
        return f"statement.Name(nm='{self.nm()}')"

    def __repr__(self) -> str:
        return f"statement.Name(nm='{self.nm()}')"



"""
[loading] section from **snowmobile.toml**, including subsections.
"""
from __future__ import annotations

from typing import Dict

from pydantic import Field

from .base import Base


class Put(Base):
    """[loading.put]"""

    auto_compress: bool = Field(default_factory=bool, alias="auto_compress")


class Copy(Base):
    """[loading.copy]"""

    on_error: str = Field(default_factory=str, alias="on_error")


# noinspection PyUnresolvedReferences
class Loading(Base):
    """[loading]

    Default settings to use when loading data

    Attributes:
        default-file-format (str):
            Name of file-format to use when loading data into the warehouse;
            default is ``snowmobile_default_csv``;
            which will be created and
            dropped afterwards if an existing file format is not specified;
        include_index (bool):
            Include the index of a DataFrame when loading it into a table;
            default is ``False``.
        on_error (bool):
            Action to take if an error is encountered when loading data into
            the warehouse; default is ``continue``.
        keep_local (bool):
            Option to keep the local file exported when loading into a
            staging table; default is ``False``.
        include_loaded_tmstmp (bool):
            Include a **loaded_tmstmp** column when loading a DataFrame into
            the warehouse; default is ``True``.
        quote_char (str):
            Quote character to use for delimited files; default is double
            quotes (``"``).
        auto_compress (bool):
            Auto-compress file when loading data; default is ``True``..
        overwrite_pre_existing_stage (bool):
            Overwrite pre-existing staging table if data is being appended
            into an existing table/the staging table already exists; default is
            ``True``.

    """

    # fmt: off
    defaults: Dict = Field(
        default_factory=dict, alias="default-table-kwargs"
    )
    put: Put = Field(
        default_factory=Put, alias="put"
    )
    copy_into: Copy = Field(
        default_factory=Copy, alias="copy-into"
    )
    export_options: Dict[str, Dict] = Field(
        default_factory=dict, alias="save-options"
    )
    # fmt: on

    @property
    def configured_args(self) -> Dict:
        """Placeholder for configuration arguments of derived classes."""
        return self.defaults



"""
Demonstrate setting attributes on the snowmobile.SQL object in place of
explicitly passing as keyword arguments.
../docs/snippets/sql_working_example2.py
"""
import snowmobile

sn1 = snowmobile.connect()
sn2 = snowmobile.connect()

for sn in [sn1, sn2]:
    sn.auto_run = False

sn1.ex("create transient table sample_table as select 1 as sample_col")

sn2.sql.nm = "sample_table"
sn2.sql.obj = "table"

# -- End Setup ----------------------------------------------------------------

sn1.sql.select(nm="sample_table")
sn2.sql.select()

sn1.sql.columns(nm="sample_table")
sn2.sql.columns()

sn1.sql.columns(nm="sample_table", from_info_schema=True)
sn2.sql.columns(from_info_schema=True)

sn1.sql.ddl(nm="sample_table")
sn2.sql.ddl()

sn1.sql.truncate(nm="sample_table")
sn2.sql.truncate()

sn1.sql.drop(nm="sample_table")
sn2.sql.drop()

# -- End Example --------------------------------------------------------------

sn2.sql.drop(run=True)

# -- End Cleanup --------------------------------------------------------------



"""
Inspect the basic parsing of a sql file.
../docs/snippets/script/intro/intro1.py
"""

# -- Setup --
from pathlib import Path
paths = {
    p.name: p
    for p in (Path.cwd() / 'docs' / 'snippets').glob('**/*.sql')
    if '.snowmobile' not in p.parts
}
path = paths['intro1.sql']

# -- Example ---

#: block 1 :
import snowmobile

script = snowmobile.Script(path=path)

#: block 2 :
script.dtl()
"""
>>>
intro1.sql
==========
1: Statement('create table~s1')
2: Statement('insert into~s2')
3: Statement('select data~s3')
"""

#: block 3 :
for i, s in script.items():
    print(f"{i}: {s}")
"""
>>>
1: Statement('create table~s1')
2: Statement('insert into~s2')
3: Statement('select data~s3')
"""

#: block 4 :
s3 = script(3)
print(s3.index)     #> 3
print(s3.sql())     #> select * from sample_table
print(s3.kw())      #> select
print(s3.anchor())  #> select data
print(s3.desc())    #> s3
print(s3.nm())      #> select data~s3


script(1).nm()
print(script(1).sql())
print(script(1).sql(tag=True))


script(1).nm()

print(script.source(original=True))
print(script.source())

script2 = snowmobile.Script(path=paths['intro1_tags.sql'])

script.depth
script.dtl()

print(script.source() == script2.source())

# -----------------------------------------------------------------------------


script(1).tag()
print(script(1).sql(tag=True))

script.dtl()

script(1).tag()
script(1)['another attribute'] = 'some value'

script(2)['index'] = 2
script(2)['name'] = script(2).nm()

print(script(3).sql(tag=True))
script(3)['some other attribute'] = 'some other piece of data'
s3_modified = script(3).sql(tag=True)

from contextlib import contextmanager


@contextmanager
def print_s(script: snowmobile.Script, n: int) -> None:
    """Print the contents of statement 'n' pre and post adding an attribute."""
    try:
        print(script(n).sql(tag=True), '\n')
        yield script
    finally:
        print(script(n).sql(tag=True))


with print_s(script, 3) as s_print:
    s_print(3)['some other attribute'] = 'some other piece of data'

snowmobile.Script(sql=s3_modified).dtl()

script(2)

print(script(1).sql(tag=True))

print(script.sn.cfg.script.tag_from_attrs(attrs=script(1).tag()))


print(script(1).sql(tag=True))
script.sn.cfg.script.markup.attrs.order

script.parse_one(script.sn.drop('sample_table', run=False))
script.dtl()

print(script.source())
print(script.source(original=True))

more_sql = """
create table any_other_table clone sample_table;

drop_table sample_table;
"""

script.parse_stream(more_sql)
script.dtl()

script.dtl()

print(script.source())

print(script.source())

print(script.source())
script(1).tag(raw=True)

script(1).anchor()
script(1).desc()

script.dtl()


script(1)['drop-by'] = '2021-12-31'
script(1).set('desc', 'sample_table')

script(1).sql()
script(1).as_section(incl_sql_tag=True).sql


script(1).run().sn.comment(
    nm=script(1).desc(),
    set_as=script(1).tag(),
    as_json=True,
    indent=4,
)

print(script.sn.comment(nm=script(1).desc(), from_json=True))

import json

sql1 = f"""
comment on table gem7318.sample_table is
'{json.dumps(script(1).tag(), indent=4)}'
"""

script.sn.query(sql1)

sql3 = script.sn.comment(script(1).desc(), run=False)
print(sql3)

script.sn.query(sql3)


detail = [script.dtl(r=True)]
with script.filter(excl_kw='select') as sf:
    detail.append(sf.dtl(r=True))
    
print('\n\n'.join(detail))

script(1).nm()


sn = snowmobile.connect()
sn.con



"""
An analytics-focused wrapper around the snowflake.connector for Python.
"""
# isort: skip_file
# fmt: off
__version__ = "0.2.0b23"
__author__ = "Grant Murray"
__application__ = "snowmobile"

__all__ = [
    # meta
    "__version__",
    "__author__",
    "__application__",

    # API
    "Snowmobile", "connect",
    "SQL",
    "Table",
    "Configuration",
    "Script",
    "Statement",
]

from .core import (
    SQL,
    Configuration,
    Snowmobile, connect,
    Table,
    Script,
    Section,
    Statement,
    utils,
)



"""
:class:`snowmobile.Script` is instantiated from a local sql file or a
readable text file containing valid sql code.
"""
from __future__ import annotations

import collections
import time
from collections import Counter
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Set,
    Any,
    Dict,
    List,
    Union,
    Tuple,
    Optional,
    ItemsView,
    KeysView,
    ValuesView,
    ContextManager,
)

import sqlparse

from . import (
    Generic,
    Diff,
    Empty,
    ExceptionHandler,
    Markup,
    Statement,
    errors,
    cfg,
)
from .connection import Snowmobile
from .cfg import Marker


class Script(Generic):
    """Parser and operator of local sql files.

    Parameters:
        sn (snowmobile.core.connection.Snowmobile):
            An instance of :class:`~snowmobile.core.connection.Snowmobile`.
        path (Optional[Path, str]):
            A full path to a sql file or readable text file containing valid
            sql code.
        path (Optional[str]):
            A raw string of valid sql code as opposed to reading from a
            ``path``.
        as_generic (bool):
            Instantiate all statements as generic st; skips all checks
            for a mapping of a statement anchor to a derived statement class
            to instantiate in the place of a generic
            :class:`~snowmobile.core.statement.Statement`.
        delay (bool):
            Delay connection of the :class:`Snowmobile`; only applicable if the
            ``sn`` argument is omitted and :class:`Script` is instantiating
            a :class:`Snowmobile` in its absence.
        **kwargs:
            Any keyword arguments to pass to :class:`Snowmobile`; only applicable
            if the ``sn`` argument is omitted and :class:`Script` is instantiating
            a :class:`Snowmobile` in its absence

    Attributes:
        sn (snowmobile.core.connection.Snowmobile):
            An instance of :class:`~snowmobile.core.connection.Snowmobile`
        patterns (snowmobile.core.cfg.script.Pattern):
            Configured patterns from :ref:`snowmobile.toml`.
        as_generic (bool):
            Instantiate all statements as generic st; skips all checks
            for a mapping of a statement anchor to a derived statement class
            to instantiate in the place of a generic
            :class:`~snowmobile.core.statement.Statement`.
        filters (Dict[Any[str, int], Dict[str, Set]]):
            Dictionary of filters that have been passed to the current instance
            of :class:`snowmobile.core.Script`.
        markers (Dict[int, cfg.Marker]):
            Dictionary of all markers found in the script.
        path (Path):
            Path to sql file (e.g. *full/path/to/script.sql*).
        name (str):
            Name of sql file (e.g. *script.sql*).
        source (str):
            Raw sql text of script; will be the text contained in the raw
            sql file when initially read from source and reflect any
            modifications to the script's contents made post-instantiation.

    """

    # Maps statement anchors to alternate base class.
    ANCHOR_TO_QA_BASE_MAP = {"qa-diff": Diff, "qa-empty": Empty}

    def __init__(
        self,
        sn: Optional[Snowmobile] = None,
        path: Optional[Path, str] = None,
        sql: Optional[str] = None,
        as_generic: bool = False,
        delay: bool = True,
        **kwargs,
    ):
        if not path and not sql:
            raise ValueError(
                "At least one of the `path` or `sql` arguments must be pr."
            )

        super().__init__()
        # TODO: Just inherit from sn
        if not sn:
            sn = Snowmobile(delay=delay, **kwargs)
        self.sn: Snowmobile = sn
        self.patterns: cfg.Pattern = sn.cfg.script.patterns
        self.as_generic = as_generic
        self.filters: Dict[Any[str, int], Dict[str, Set]] = {
            int(): {k: v for k, v in self.sn.cfg.scopes.items() if v}
        }
        self.filtered: bool = bool()
        self.markers: Dict[int, cfg.Marker] = dict()
        self.path: Path = Path()
        self.name: str = str()
        self._source: str = str()
        self._source_end: int = int()

        self._is_from_str: Optional[bool] = None
        self._is_post_init: bool = False
        self._statements_all: Dict[int, Statement] = dict()
        self._statements_parsed: Dict[int, sqlparse.sql.Statement] = dict()
        self._open, self._close = sn.cfg.script.tag()
        self._intra_statement_marker_hashmap_idx: Dict = dict()
        self._intra_statement_marker_hashmap_txt: Dict = dict()
        self._all_marker_hashmap: Dict = dict()

        if path and not sql:
            try:
                self.read(path=path)
            except IOError as e:
                raise e
        elif sql:
            is_pr, _pr = bool(path), Path(str(path))
            _nm = 'snowmobile.sql' if not is_pr else _pr.name  # TODO: Refine
            _dir = _pr.parent if is_pr else Path.cwd()
            self.from_str(
                sql=sql,
                name=_nm,
                directory=_dir
            )

        self.e = ExceptionHandler(
            within=self,
            children=self.st,
            is_active_parent=True,
            to_mirror=["set", "reset"],
        )

        self._stdout: Script.Stdout = self.Stdout(name=self.name, statements=dict())

    def _post_source__init__(self, from_str: bool = False) -> Script:
        """Sets final attributes and parses source once pr."""
        self.name = self.path.name
        self._is_from_str = from_str
        self._source_end = len(self._source.split('\n'))

        try:
            self._parse()
        except Exception as e:
            raise e

        self._is_post_init = True
        return self

    def read(self, path: Path = None) -> Script:
        """Runs quick path validation and reads in a sql file as a string.

        A valid `path` must be pr if the `script.path` attribute hasn't
        been set; ``ValueErrors`` will be thrown if neither is valid.

        Args:
            path (pathlib.Path):
                Full path to a sql object.

        """
        try:
            self.path: Path = Path(str(path)) if path else self.path
            self.name = self.path.name
            with open(self.path, "r") as r:
                self._source = r.read()

            return self._post_source__init__()

        except IOError as e:
            raise e

    def from_str(self, sql: str, name: str, directory: Path = Path.cwd()) -> Script:
        """Instantiates a raw string of sql as a script."""
        # fmt: off
        if not name.endswith(".sql"):
            raise ValueError(
                f"`name` must end in .sql; '{name}' pr."
            )
        # fmt: on
        self._source = sql
        self.path: Path = Path(str(directory)) / name
        return self._post_source__init__(from_str=True)

    def source(self, original: bool = False) -> str:
        """The script's sql as a raw string."""
        if original:
            return '\n'.join(self._source.split('\n')[0:self._source_end])
        return '\n\n'.join(f"{s.sql(tag=True)};" for _, s in self.items())

    def _stream(self, stream: Optional[str] = None) -> sqlparse.sql.Statement:
        """Parses source sql into individual st."""
        for s in sqlparse.parsestream(stream=stream or self._source):
            if self.sn.cfg.script.is_valid_sql(s=s):
                yield s

    def parse_one(
        self,
        s: Union[sqlparse.sql.Statement, str],
        index: Optional[int] = None,
        nm: Optional[str] = None,
    ) -> None:
        """Adds a statement object to the script.

        Default behavior will only add ``sqlparse.sql.Statement`` objects
        returned from ``script.source_stream``.

        ``clean_parse()`` utility function is utilized so that ge sql
        within Python can be inserted back into the script as raw strings.

        Args:
            s (Union[sqlparse.sql.Statement, str]):
                A sqlparse.sql.Statement object or a raw string of SQL for an
                individual statement.
            index (int):
                Index position of the statement within the script; defaults
                to ``n + 1`` if index is not pr where ``n`` is the number
                of statements within the script at the time ``parse_one()``
                is called.
            nm (Optional[str]):
                Optionally pr the name of the statement being added; the
                script instance will treat this value as if it were pr
                within an in-script wrap.

        """
        index = index or self.depth + 1
        if nm:
            s: str = f"{self._open}{nm}{self._close}\n{s}"
        s: sqlparse.sql.Statement = self.sn.cfg.script.ensure_sqlparse(sql=s)

        markers, attrs_raw = self.sn.cfg.script.split_sub_blocks(s=s)
        self._log_markers(idx=index, markers=markers)
        self._add_s(s=s, index=index, attrs_raw=attrs_raw)
        
    def parse_stream(self, stream: str) -> None:
        """Parses a stream of sql and adds onto existing Script contents."""
        self._parse_statements(stream=stream)

    def _add_s(
        self, s: sqlparse.sql.Statement, index: int, attrs_raw: str
    ) -> None:
        """Adds a statement object to the script.

        Instantiates a generic statement object, immediately stores by index if
        it's anchor indicates that it's not intended to be a QA statement,
        otherwise instantiates the associated QA statement and stores that
        instead.

        Args:
            s (sqlparse.sql.Statement):
                sqlparse Statement object.
            index (int):
                Statement's index position within script.
            attrs_raw (str):
                Raw string of attributes parsed from the statement.

        """
        # generic case
        statement: Any[Statement, Empty, Diff] = Statement(
            sn=self.sn, statement=s, index=index, attrs_raw=attrs_raw
        )
        if not statement.is_derived or self.as_generic:
            self._statements_all[index] = statement
        # mapping to associated QA base class otherwise
        else:
            self._statements_all[index] = self._derive_qa_from_generic(
                s=s, generic=statement
            )

        # method is being invoked by user, not initial object instantiation
        if self._is_post_init:
            self._source = f"{self._source}\n{self._statements_all[index].trim()}"

    def _log_markers(self, idx: int, markers: List[str]) -> None:
        """Stores intra-statement markers.

        Args:
            idx (int):
                Statement index.
            markers (List[str]):
                List of markers as raw strings found above the statement.
        """
        for i, m in enumerate(markers, start=1):
            _hash = hash(m)
            marker_index = idx + (i / 10)
            self._intra_statement_marker_hashmap_idx[_hash] = marker_index
            self._intra_statement_marker_hashmap_txt[_hash] = m

    def _derive_qa_from_generic(
        self, s: sqlparse.sql.Statement, generic: Statement
    ) -> Union[Diff, Empty]:
        """Instantiates a QA statement object based off the statement's anchor."""
        qa_base_class = self.ANCHOR_TO_QA_BASE_MAP[generic.anchor(pr=True)]
        return qa_base_class(
            sn=self.sn,
            statement=s,
            index=generic.index,
            attrs_raw=generic.tag(raw=True)
        )

    def _parse_statements(self, stream: Optional[str] = None) -> None:
        """Instantiates statement objects for all statements in source sql."""
        if not stream:
            self._statements_all.clear()
        parsed = self._stream(stream=stream)
        for i, s in enumerate(parsed, start=self.depth + 1):
            self.parse_one(s=s, index=i)

    def _parse_markers(self):
        """Parses all markers into a hashmap to the raw text within the marker."""
        cfg = self.sn.cfg.script
        self._all_marker_hashmap = {
            hash(m): m
            for m in cfg.find_tags(sql=self._source).values()
            if cfg.is_marker(m)
        }

    def _parse(self):
        """Parses statements and markers within :attr:`source`."""
        try:
            self._parse_statements()
            self._parse_markers()
            self.markers = self.sn.cfg.script.markup.attrs.merge_markers(
                parsed_markers=self._parsed_markers
            )
        except Exception as e:
            raise e

    def _scope_from_id(self, _id: Any[int, str], pop: bool = True) -> Dict:
        """Returns dictionary of scope arguments given an ``_id``.

        Will return scope arguments from ``script.filters`` if ``_id`` is
        pre-existing and template value if not.

        Args:
            _id (Any[int, str]):
                Integer or string value for scope ``ID``.
            pop (bool):
                Remove scope from ``script.scope`` before returning;
                default=False.

        Returns (dict):
            A dictionary of scope keys to associated values.

        """
        template = self.sn.cfg.scopes
        if _id not in self.filters:
            return template
        else:
            scope = self.filters.pop(_id) if pop else self.filters[_id]
        return {k: scope.get(k, template[k]) for k in template}

    def _update_scope_statements(self, scope_to_set: Dict) -> None:
        """Applies a set of scope arguments to all statements within the script.

        Args:
            scope_to_set (dict):
                A full set of scope arguments.
        """
        for s in self._statements_all.values():
            s.set_state(ctx_id=self.e.ctx_id, in_context=True, filters=scope_to_set)

    def _update_scope_script(self, _id: Any[int, str], **kwargs) -> Dict:
        """Returns a valid set of scope args from an ``_id`` and the scope kwargs.

        Uses template property from configuration if the ``_id`` pr does
        not yet exist in ``script.filters``.

        Args:
            _id (Any[int, str]):
                Integer or string value for scope _id.
            **kwargs:
                Arguments pr to ``script.filter()`` (e.g. 'include_kw',
                'excl_anchor', etc).

        """
        _id = _id or (len(self.filters) + 1)
        # TODO: Remove this argument/feature in script.filter(); super not worth
        #   the time.
        scope_to_update = (
            self._scope_from_id(_id=_id, pop=True)
            if _id in self.filters
            else self.sn.cfg.scopes
        )
        merged_with_latest_filter = {
            arg: filters.union(kwargs[arg]) for arg, filters in scope_to_update.items()
        }
        self.filters[_id] = {k: v for k, v in merged_with_latest_filter.items() if v}
        return self.filters[_id]

    # DOCSTRING
    def _update_scope(
        self,
        as_id: Optional[Union[int, str]],
        from_id: Optional[Union[int, str]],
        **kwargs,
    ):
        _id = from_id or as_id
        if from_id:
            scope_config = self._scope_from_id(_id=_id, pop=False)
        else:
            scope_config = self.sn.cfg.scopes_from_kwargs(**kwargs)
        latest_scope = self._update_scope_script(_id=_id, **scope_config)
        self._update_scope_statements(scope_to_set=latest_scope)

    # DOCSTRING
    @contextmanager
    def filter(
        self,
        incl_kw: Optional[List[str], str] = None,
        incl_obj: Optional[List[str], str] = None,
        incl_desc: Optional[List[str], str] = None,
        incl_anchor: Optional[List[str], str] = None,
        incl_nm: Optional[List[str], str] = None,
        excl_kw: Optional[List[str], str] = None,
        excl_obj: Optional[List[str], str] = None,
        excl_desc: Optional[List[str], str] = None,
        excl_anchor: Optional[List[str], str] = None,
        excl_nm: Optional[List[str], str] = None,
        as_id: Optional[Union[str, int]] = None,
        from_id: Optional[Union[str, int]] = None,
        last: bool = False,
    ) -> ContextManager[Script]:
        """Subset the script based on attributes of its st.

        ``script.filter()`` returns a modified instance of script that can
        be operated on within the context defined.

        .. note::
            Keyword arguments beginning with ``incl`` or ``excl`` expect a
            string or a list of strings containing regex patterns with which
            to check for a match against the associated attribute of its
            st' :class:`~snowmobile.core.name.Name`.

        Args:
            incl_kw:
                Include only :attr:`~snowmobile.core.name.Name.kw`
            incl_obj:
                Include only :attr:`~snowmobile.core.name.Name.obj`
            incl_desc:
                Include only :attr:`~snowmobile.core.name.Name.desc`
            incl_anchor:
                Include only :attr:`~snowmobile.core.name.Name.anchor`
            incl_nm:
                Include only :attr:`~snowmobile.core.name.Name.nm`
            excl_kw:
                Exclude :attr:`~snowmobile.core.name.Name.kw`
            excl_obj:
                Exclude :attr:`~snowmobile.core.name.Name.obj`
            excl_desc:
                Exclude :attr:`~snowmobile.core.name.Name.desc`
            excl_anchor:
                Exclude :attr:`~snowmobile.core.name.Name.anchor`
            excl_nm:
                Exclude :attr:`~snowmobile.core.name.Name.nm`
            as_id:
                ID to assign the filters passed to method; used to populated
                the :attr:`filters` attribute
            from_id:
                ID previously used on the same instance of :class:`Script`
                from which to populate filtered arguments
            last:
                Re-use the last set of filters passed to context manager.

        Returns (Script):
            The instance of script based on the context imposed by arguments
            pr.

        """
        def to_list(arg: Union[List, str, None]) -> Union[None, List]:
            """Enable passing a list or a string to filter arguments."""
            if not arg:
                return arg
            return arg if isinstance(arg, List) else [arg]

        # fmt: off
        if from_id and as_id:
            raise ValueError(
                f"script.filter() cannot accept `from_id` and `as_id` arguments"
                f" simultaneously."
            )
        if from_id and from_id not in self.filters:
            raise ValueError(
                f"from_id='{from_id}' does not exist in script.filters;"
                f"IDs that do exist are: {list(self.filters.keys())}"
            )

        try:
            self.e.set(ctx_id=time.time_ns(), in_context=True)

            if last:
                from_id, as_id = list(self.filters)[-1], None

            filters = {
                'incl_kw': incl_kw,
                'incl_obj': incl_obj,
                'incl_desc': incl_desc,
                'incl_anchor': incl_anchor,
                'incl_nm': incl_nm,
                'excl_kw': excl_kw,
                'excl_obj': excl_obj,
                'excl_desc': excl_desc,
                'excl_anchor': excl_anchor,
                'excl_nm': excl_nm,

            }
            filters_adj = {k: to_list(v) for k, v in filters.items()}
            self._update_scope(
                as_id=as_id,
                from_id=from_id,
                last=last,
                **filters_adj,
            )

            yield self.reset(_filter=True)  # script.filtered = True; filter imposed

        except Exception as e:
            self.e.collect(e=e)

        finally:

            # first collect e
            to_raise = (
                self.e.get(last=True, to_raise=True)
                if self.e.seen(to_raise=True)
                else None
            )

            # then reset context
            self.reset(
                index=True,       # restore statement indices
                scope=True,       # reset included/excluded status of all statements
                ctx_id=True,      # cache context tmstmp for both script and statements
                in_context=True,  # release 'in context manager' indicator (to False)
                _filter=True,     # release 'impose filter' indicator (to  False)
            )

            # then give e his due
            if to_raise:
                raise to_raise

            # fmt: on
            return self

    def _depth(self, full: bool = False) -> int:
        return len(self._adjusted_contents) if full else len(self.st)

    @property
    def depth(self) -> int:
        """Count of statements in the script."""
        if not self._statements_all:
            return 0
        return self._depth()

    @property
    def lines(self) -> int:
        """Number of lines in the script"""
        return sum(len(s) for s in self.st.values())

    def _id(self, _id: Union[int, str]) -> int:
        """Returns index position of a statement given its index or wrap name."""
        if isinstance(_id, int):
            return _id if _id > 0 else (self.depth + _id + 1)
        try:
            statements = {
                k: v
                for k, v in self.items(
                    by_index=False,
                    validate=bool(_id in self.duplicates),
                )
            }
            return statements[_id].index
        except Exception as e:
            raise e

    @property
    def excluded(self):
        """All statements by index position excluded from the current context."""
        return {
            i: s for i, s in self._statements_all.items() if i not in self.st
        }

    @property
    def executed(self) -> Dict[int, Statement]:
        """Executed statements by index position included in the current context."""
        return {i: s for i, s in self.st.items() if s.executed}

    def reset(
        self,
        index: bool = False,
        ctx_id: bool = False,
        in_context: bool = False,
        scope: bool = False,
        _filter: bool = False,
    ) -> Script:
        """Resets indices and scope on all statements to their state as read from source.

        Invoked before exiting :meth:`filter()` context manger to reverse
        the revised indices set by :meth:`index_to()` and inclusion/
        exclusion scope set by :meth:`Statement.Name.scope()`.

        """

        def batch_reset(**kwargs) -> Dict[Statement]:
            """Calls .reset() with kwargs on all statement objects."""
            return {i: s.reset(**kwargs) for i, s in self._statements_all.items()}

        if _filter:  # NOTE: must come before re-index
            self.filtered = not bool(self.filtered)
        if index:
            re_indexed = batch_reset(index=index)
            unsorted_by_index = {s.index: s for s in re_indexed.values()}
            self._statements_all = {
                i: unsorted_by_index[i] for i in sorted(unsorted_by_index)
            }
        if ctx_id:
            self.e.reset(ctx_id=True)
        if in_context:
            self.e.set(in_context=False)
        if scope:
            self._statements_all = batch_reset(scope=scope)

        return self

    @property
    def duplicates(self) -> Dict[str, int]:
        """Dictionary of indistinct statement names/tags within script."""
        counted = collections.Counter([s.nm() for s in self._statements_all.values()])
        return {tag: cnt for tag, cnt in counted.items() if cnt > 1}

    def s(self, _id: Optional[str, int] = None) -> Any[Statement, Empty, Diff]:
        """Fetch a single statement by _id."""
        index_of_id = self._id(_id=_id)
        if index_of_id not in self.st:
            raise errors.StatementNotFoundError(nm=_id)
        return self.st[index_of_id]

    @property
    def st(self) -> Dict[Union[int, str], Statement]:
        """Accessor for all statements."""
        if not self.filtered:
            return self._statements_all
        statements = {s.index: s for s in self._statements_all.values() if s}
        return {
            current_idx: statements[prior_idx].set_state(index=current_idx)
            for current_idx, prior_idx in enumerate(sorted(statements), start=1)
        }

    def dtl(
        self,
        full: bool = False,
        excluded: bool = False,
        title: bool = True,
        r: bool = False,
    ) -> Union[str, None]:
        """Prints summary of statements within the current scope to console."""
        dtl = self.st if not full else self._adjusted_contents
        depth = self._depth(full=full)
        if excluded:
            dtl, depth = self.excluded, len(self.excluded)
        _text = []
        if title:
            _text.append(self._console.display(r=True))
        for i, s in dtl.items():
            _text.append(f"{str(i).rjust(len(str(depth)), ' ')}: {s}")
        text = '\n'.join(_text)
        return self._console.cprint(text, r)

    @property
    def first_s(self):
        """First statement by index position."""
        return self.st[min(self.st)]

    @property
    def last_s(self):
        """Last statement by index position"""
        return self.st[max(self.st)]

    @property
    def first(self) -> Union[Statement, Empty, Diff]:
        """First statement executed."""
        by_start = {v.start_time: v for v in self.executed.values()}
        return by_start[min(by_start)] if by_start else None

    @property
    def last(self) -> Union[Statement, Empty, Diff]:
        """Last statement executed."""
        by_start = {v.start_time: v for v in self.executed.values()}
        return by_start[max(by_start)] if by_start else None

    def doc(
        self,
        nm: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        incl_markers: Optional[bool] = True,
        incl_sql: Optional[bool] = True,
        incl_exp_ctx: Optional[bool] = True,
        result_wrap: Optional[str] = None,
    ) -> Markup:
        """Returns a :class:`~snowmobile.core.markup.Markup` from the script.

        Args:
            nm (Optional[str]):
                Alternate file name to use.
            prefix (Optional[str]):
                Prefix for file name.
            suffix (Optional[str]):
                Suffix for file name.
            incl_markers (Optional[bool]):
                Include markers in exported files.
            incl_sql (Optional[bool]):
                Include sql in exported files.
            incl_exp_ctx (Optional[bool]):
                Include disclaimer of programmatic save in exported sql file.

        Returns:
            A :class:`~snowmobile.core.markup.Markup` instance based on the
            contents included in the script's context.

        """
        return Markup(
            sn=self.sn,
            path=self.path,
            contents=self._adjusted_contents,
            nm=nm,
            prefix=prefix,
            suffix=suffix,
            incl_sql=incl_sql,
            incl_markers=incl_markers,
            incl_exp_ctx=incl_exp_ctx,
            result_wrap=result_wrap,
        )

    @property
    def _intra_statement_markers(self):
        """All markers (raw text) above/between statements by index position."""
        return {
            i: self._intra_statement_marker_hashmap_txt[h]
            for h, i in self._intra_statement_marker_hashmap_idx.items()
        }

    @property
    def _trailing_statement_markers(self) -> Dict[float, str]:
        """All markers (raw text) after the last statement by index position."""
        markers_r_unadjusted = {
            m
            for h, m in self._all_marker_hashmap.items()
            if not self._intra_statement_marker_hashmap_idx.get(h)
        }
        return {
            (self.depth + 1 + (i / 10)): m
            for i, m in enumerate(markers_r_unadjusted, start=1)
        }

    @property
    def _ordered_markers(self) -> Dict[int, str]:
        """All markers as raw text, ordered by index position."""
        all_markers = {
            **self._intra_statement_markers,
            **self._trailing_statement_markers,
        }
        return {i: all_markers[i] for i in sorted(all_markers)}

    @property
    def _parsed_markers(self):
        """All markers (as dictionaries), ordered by index position."""
        return {
            i: self.sn.cfg.script.parse_marker(attrs_raw=block)
            for i, block in self._ordered_markers.items()
        }

    @property
    def _marker_counter(self):
        """Dictionary of the number of markers at each statement index."""
        return collections.Counter(round(k) for k in self.markers)

    def _get_marker_adjusted_idx(
        self, idx: int, counter: Counter, is_marker: bool = False, as_int: bool = False
    ) -> int:
        """Generates an index (int) taking into account statements and markers."""
        # sourcery skip: simplify-constant-sum
        index = idx + (
            sum(v for k, v in counter.items() if k <= idx)
            if not is_marker
            else sum(1 for k in self.markers.keys() if k < idx)
        )
        return index if not as_int else round(index)

    def _adjusted_statements(self, counter: Counter):
        """Statements by adjusted index position."""
        return {
            self._get_marker_adjusted_idx(idx=i, counter=counter): s
            for i, s in self.st.items()
        }

    def _adjusted_markers(self, counter: Counter):
        """Markers by adjusted index position."""
        adjusted_markers = {}
        for i, m in self.markers.items():
            i_adj = self._get_marker_adjusted_idx(
                idx=i, counter=counter, is_marker=True, as_int=True
            )
            adjusted_markers[i_adj] = m
        return adjusted_markers

    @property
    def _adjusted_contents(self):
        """All statements and markers by adjusted index position."""
        try:
            counter = Counter(round(k) for k in self.markers)
            adjusted_statements = self._adjusted_statements(counter=counter)
            adjusted_markers = self._adjusted_markers(counter=counter)
            contents = {**adjusted_markers, **adjusted_statements}
            return {i: contents[i] for i in sorted(contents)}
        except Exception as e:
            raise e

    def ids(self, _id: Optional[Union[Tuple, List]] = None) -> List[int]:
        """Utility function to get a list of statement IDs given an `_id`.

        Invoked within script.run() if the `_id` parameter is either a:
            (1) tuple of integers (lower and upper bound of statement indices
                to run)
            (2) list of integers or strings (statement names or indices to run)
            (3) default=None; returns all statement indices within scope if so

        Args:
            _id (Union[Tuple, List]):
                _id field pr to script.run() if it's neither an integer
                or a string.

        Returns (List[int]):
            A list of statement indices to run.

        """
        if not _id:
            return list(self.st)
        if isinstance(_id, List):
            return _id
        elif isinstance(_id, Tuple):
            start_intl, stop_intl = _id
            start, stop = (
                (self._id(start_intl) if start_intl else 1),
                (self._id(_id=stop_intl) + 1),
            )
            return [i for i in range(start, stop)]

    def run(
        self,
        _id: Optional[str, int, Tuple[int, int], List] = None,
        as_df: bool = True,
        on_error: Optional[str] = None,
        on_exception: Optional[str] = None,
        on_failure: Optional[str] = None,
        lower: bool = True,
        render: bool = False,
        **kwargs,
    ) -> None:
        """Performs statement-by-statement execution of the script's contents.

        Executes script's contents that are included within its current context
        and any (optional) value passed to the ``_id`` argument.

        .. note::
            Keyword arguments ``on_exception`` and ``on_failure`` are only
            applicable to derived classes of
            :class:`~snowmobile.core.statement.Statement`
            (e.g., those within :mod:`snowmobile.core.qa` by default).

        Args:
            _id (Optional[str, int, Tuple[int, int], List]):
                Identifier for statement(s) to execute, can be either:
                    - *None* (default); execute all statements
                    - A single statement's :attr:`~snowmobile.core.Name.nm`
                    - A single statement's index position
                    - A tuple of lower/upper index bounds of statements to execute
                    - A list of statement names or index positions to execute
            as_df (bool):
                Store statement's results as a :class:`~pandas.DataFrame`;
                defaults to *True*
            on_error (Optional[str]):
                Action to take on **execution** error; providing `c` will
                continue execution as opposed to raising exception.
            on_exception (Optional[str]):
                Action to take on **post-processing** error from a derived
                :class:`~snowmobile.core.statement.Statement`; providing `c`
                will continue execution as opposed to raising exception.
            on_failure (Optional[str]):
                Action to take on **failure** of post-processing assertion from
                a derived :class:`~snowmobile.core.statement.Statement`;
                providing `c` will continue execution as opposed to raising
                exception.
            lower (bool):
                Lower-case columns in results returned if ``as_df=True``.
            render (bool):
                Render sql executed as markdown; only applicable in
                Jupyter/iPython environments.
            **kwargs:
        """
        def _run(_id: Optional[str, int] = None, v: bool = True, **kwargs):
            """Fetches a statement, runs it, prints outcome to console."""
            s = self.s(_id)
            s.run(**kwargs)
            if v:
                self._console.status(s)

        if not self.e.in_context:
            self.e.set(ctx_id=-1)
        total_kwargs = {
            **{
                "as_df": as_df,
                "on_error": on_error,
                "on_exception": on_exception,
                "on_failure": on_failure,
                "lower": lower,
                "render": render,
            },
            **kwargs,
        }
        if isinstance(_id, (int, str)):
            _run(_id=_id, v=not render, **total_kwargs)
        else:
            indices_to_execute = self. ids(_id=_id)
            self._console.display()
            for i in indices_to_execute:
                _run(_id=i, v=not render, **total_kwargs)
    
    def items(
        self,
        by_index: bool = True,
        ignore_scope: bool = False,
        statements: bool = True,
        markers: bool = False,
        validate: bool = True,
    ) -> ItemsView[Union[int, str], Union[Statement, Marker]]:
        """Dunder items."""
        if by_index and statements and not markers and not ignore_scope:
            return self.st.items()
        elif by_index and statements and not markers:
            return self._statements_all.items()
        elif statements and not markers:
            return {s.nm(): s for s in self.st.values()}.items()
        contents_to_return = self._adjusted_contents
        if not statements:
            contents_to_return = {
                i: s
                for i, s in contents_to_return.items()
                if not any(
                    [isinstance(s, Statement), issubclass(Statement, type(s))]
                )
            }
        if by_index:
            return contents_to_return.items()
        # validation to ensure keys are unique if fetching contents by wrap name
        if validate and not (
            len({s for s in contents_to_return})
            == len({s.nm() for s in contents_to_return.values()})
        ):
            raise errors.DuplicateTagError(nm=self.path.name)
        return {s.nm(): s for i, s in contents_to_return.items()}.items()
    
    def keys(self, **kwargs) -> KeysView[Union[int, str]]:
        """Access keys of items only."""
        return {k: v for k, v in self.items(**kwargs)}.keys()
    
    def values(self, **kwargs) -> ValuesView[Union[int, str]]:
        """Access values of items only."""
        return {k: v for k, v in self.items(**kwargs)}.values()
    
    def dict(self, **kwargs) -> Dict:
        """Unpacking items view into an actual dictionary."""
        return {k: v for k, v in self.items(**kwargs)}

    def __call__(self, _id: Union[int, str]) -> Statement:
        return self.s(_id=_id)

    def __str__(self) -> str:
        return f"snowmobile.Script('{self.name}')"

    def __repr__(self) -> str:
        return f"snowmobile.Script('{self.name}')"

    def __iter__(self):
        """Dunder iteration."""
        return self.st

    @property
    def _console(self):
        """External stdout object for console feedback without cluttering code."""
        self._stdout.statements = self.st
        return self._stdout

    # noinspection PyMissingOrEmptyDocstring
    class Stdout:
        """Console output."""

        def __init__(
            self, name: str, statements: Dict[int, Statement], verbose: bool = True
        ):
            self.name: str = name
            self.statements = statements
            self.verbose = verbose
            self.max_width_outcome = len("<COMPLETED>")
            self.outputs: Dict[int, str] = {}

        @property
        def cnt_statements(self) -> int:
            return len(self.statements)

        @property
        def max_width_progress(self) -> int:
            return max(
                len(f"<{i} of {self.cnt_statements}>")
                for i, _ in enumerate(self.statements.values(), start=1)
            )

        @property
        def max_width_tag_and_time(self) -> int:
            return max(len(f"{s.nm()} (~0s)") for s in self.statements.values())

        def console_progress(self, s: Statement) -> str:
            return f"<{s.index} of {self.cnt_statements}>".rjust(
                self.max_width_progress, " "
            )

        def console_tag_and_time(self, s: Statement) -> str:
            return f"{s.nm()} ({s.execution_time_txt})".ljust(
                self.max_width_tag_and_time + 3, "."
            )

        def console_outcome(self, s: Statement) -> str:
            return f"<{s.outcome_txt().lower()}>".ljust(self.max_width_outcome, " ")

        def status(self, s: Statement, r: bool = False) -> Union[None, str]:
            progress = self.console_progress(s)
            tag_and_time = self.console_tag_and_time(s)
            outcome = self.console_outcome(s)
            stdout = f"{progress} {tag_and_time} {outcome}"
            self.outputs[s.index] = stdout
            return self.cprint(stdout, r)

        def display(self, underline: bool = True, r: bool = False) -> Union[str, None]:
            name = self.name
            if underline:
                bottom_border = "=" * len(name)
                name = f"{name}\n{bottom_border}"
            return self.cprint(name, r)
            
        def cprint(self, text: str, r: bool) -> Union[str, None]:
            """Conditionally print or return 'text' based on 'r'."""
            if self.verbose and not r:
                print(text)
            if r:
                return text



"""Tests for :class:`snowmobile.Script` functionality."""
import pytest

from test import FILES

QA_SCRIPT_NAME = "tags_qa_statements.sql"


@pytest.fixture(scope="session")
def setup_for_qa_statement_tests(sn):
    """Sets up script and temp table for QA statement tests."""
    from snowmobile import Script

    script = Script(path=FILES[QA_SCRIPT_NAME], sn=sn)
    script.run(1)  # create temp table for st to run on
    return script


@pytest.mark.qa
def test_qa_empty_passes(setup_for_qa_statement_tests):
    """Verifies that a valid `qa-empty` passes validation and contains correct attributes.
    """
    script = setup_for_qa_statement_tests
    with script.filter(incl_anchor=["qa-empty"], incl_desc=[".*expected pass"]) as s:
        s.run()  # run w/ default error handling
        for s2 in s.executed.values():
            assert s2.outcome  # outcome bool
            assert s2.outcome_txt() == "passed"  # outcome text
            assert not s2.e  # no errors encountered


@pytest.mark.qa
def test_qa_diff_passes(setup_for_qa_statement_tests):
    """Verifies that a valid `qa-empty` throws no exception and contains the
    correct attributes.
    """
    script = setup_for_qa_statement_tests
    with script.filter(incl_anchor=["qa-diff"], incl_desc=[".*expected pass"]) as s:
        s.run()  # run w/ default error handling
        for s2 in s.executed.values():
            assert s2.outcome  # outcome bool
            assert s2.outcome_txt() == "passed"  # outcome text
            assert not s2.e  # no errors encountered


@pytest.mark.qa
def test_qa_empty_failures(setup_for_qa_statement_tests):
    """Verifies that a valid `qa-empty` throws no exception and contains the
    correct attributes.
    """
    script = setup_for_qa_statement_tests
    with script.filter(incl_anchor=["qa-empty"], incl_desc=[".*expected failure"]) as s:
        s.run(on_failure="c")  # SILENCING QA FAILURES
        for s2 in s.executed.values():
            assert not s2.outcome  # NOT outcome bool
            assert s2.outcome_txt() == "failed"  # 'failed' outcome text
            assert s2.e  # errors encountered


@pytest.mark.qa
def test_qa_diff_failures(setup_for_qa_statement_tests):
    """Verifies that a valid `qa-empty` throws no exception and contains the
    correct attributes.
    """
    script = setup_for_qa_statement_tests
    with script.filter(incl_anchor=["qa-diff"], incl_desc=[".*expected failure"]) as s:
        s.run(on_failure="c")  # SILENCING QA FAILURES
        for s2 in s.executed.values():
            assert not s2.outcome  # NOT outcome bool
            assert s2.outcome_txt() == "failed"  # 'failed' outcome text
            assert s2.e  # errors encountered


@pytest.mark.exceptions
@pytest.mark.qa
def test_qa_empty_failure_exceptions(setup_for_qa_statement_tests):
    """Verifies that an invalid `qa-empty` throws appropriate exception."""
    from snowmobile.core.errors import QAEmptyFailure

    script = setup_for_qa_statement_tests
    with script.filter(incl_anchor=["qa-empty"], incl_desc=[".*expected failure"]) as s:

        with pytest.raises(QAEmptyFailure):
            s.run()
            # try:
            #     s.run()         # verify with `script.run()`
            # except Exception as e:
            #     to_inspect = e
            #     raise e
        for i in s.st:
            with pytest.raises(QAEmptyFailure):
                s.run(i)
                # try:
                #     s.run(i)    # verify when manually calling `script.run()` on each statement
                # except Exception as e:
                #     to_inspect2 = e
                #     raise e
            with pytest.raises(QAEmptyFailure):
                s(i).run()
                # try:
                #     s(i).run()  # verify when calling each individual statement's `.run()` method
                # except Exception as e:
                #     to_inspect3 = e
                #     raise e


@pytest.mark.exceptions
@pytest.mark.qa
def test_qa_diff_failure_exceptions(setup_for_qa_statement_tests):
    """Verifies that an invalid `qa-diff` throws appropriate exception."""
    from snowmobile.core.errors import QADiffFailure

    script = setup_for_qa_statement_tests
    with script.filter(incl_anchor=["qa-diff"], incl_desc=[".*expected failure"]) as s:
        with pytest.raises(QADiffFailure):
            s.run()

        for i in s.st:
            with pytest.raises(QADiffFailure):
                s.run(i)
            with pytest.raises(QADiffFailure):
                s(i).run()


@pytest.mark.exceptions
@pytest.mark.qa
def test_qa_diff_post_processing_exception(setup_for_qa_statement_tests):
    """Verifies that an invalid `qa-diff` throws appropriate exception."""
    from snowmobile.core.errors import StatementPostProcessingError

    script = setup_for_qa_statement_tests

    with script.filter(incl_nm=["verify an exception is thrown"]) as s:

        with pytest.raises(StatementPostProcessingError):
            s.run()

        for i in s.st:
            with pytest.raises(StatementPostProcessingError):
                s.run(i)

            with pytest.raises(StatementPostProcessingError):
                s(i).run()



""" Establish a basic connection.
../docs/snippets/connecting.py
"""
import snowmobile

sn = snowmobile.connect()

print(sn)            #> snowmobile.Snowmobile(creds='creds1')
print(sn.cfg)        #> snowmobile.Configuration('snowmobile.toml')
print(type(sn.con))  #> <class 'snowflake.connector.connection.SnowflakeConnection'>

sn2 = snowmobile.connect(creds="creds1")

sn.cfg.connection.current == sn2.cfg.connection.current  #> True
sn.current("schema") == sn2.sql.current("schema")    #> True
sn.current("session") == sn2.sql.current("session")  #> False

# -- complete example; should run 'as is' --
# snowmobile-include



# import link
from . import link

link_name = "copy into table"
user_text = "copy into table"
url = "https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html"

link.xref_links.update({link_name: (user_text, url)})



"""
This file stores setup and tests for snowmobile.Markup.

See `context.md` within `/tests/func/data/sql` for more information on the
directory structure/approach to this part of the testing.
"""
import pytest

from typing import Dict

from test import script as get_script, contents_are_identical, get_validation_file

import snowmobile


# noinspection PyProtectedMember
def export_markup_combinations_from_script(
    script: snowmobile.Script, run: bool, run_args: Dict = None
):
    """Exports different versions of a `Script` object given sets of args.

    Args:
        script (Script):
            Script object under test.
        run (bool):
            Indicates whether or not to run the script before exporting.
        run_args (dict):
            Arguments to pass to `script.run()` if `run=True`

    Returns (List[Path]):
        A list of the paths that were exported which is used within the body
        of the test method to fetch the validation file and assert their equality.

    export_args (explanation)
        *   config:
                Arguments to pass to markup.config() or to the markup object as
                a callable; i.e. markup()(**args) as opposed to
                markup.config(**args)
        *   save:
                Arguments to pass to markup.save() to modify the contents
                that are exported and the target file name.

    """

    export_args = [
        {"config": {"prefix": "(default) "}, "save": {}},
        {"config": {"prefix": "(no sql) "}, "save": {"sql": False}},
        {
            "config": {"incl_markers": False, "prefix": "(no markers) "},
            "save": {},
        },
        {
            "config": {
                "incl_markers": False,
                "incl_exp_ctx": False,
                "prefix": "(no disclaimer, no markers) ",
            },
            "save": {"md": False},
        },
    ]

    if run:
        # excluding execution time from unit test as this can vary run to run
        script.sn.cfg.attrs.exclude("execution_time_txt")
        script.run(**run_args or dict())
    file_paths = []
    for arg in export_args:
        markup = script.doc()(**arg["config"])
        markup.save(**arg["save"])
        for p in markup.exported:
            file_paths.append(p)
    return file_paths


@pytest.mark.markup
def test_markup_with_results(sn):
    """Unit test for markup save including as_df."""
    script = get_script("markup_with_results.sql")

    exported_paths_for_current_test = export_markup_combinations_from_script(
        script=script, run=True, run_args={"on_failure": "c", "on_exception": "c"}
    )
    paths_under_test_mapped_to_validation_paths = {
        p: get_validation_file(path1=p) for p in exported_paths_for_current_test
    }

    for (
        p_under_test,
        p_validation,
    ) in paths_under_test_mapped_to_validation_paths.items():
        assert contents_are_identical(path1=p_under_test, path2=p_validation)


@pytest.mark.markup
def test_markup_no_results(sn):
    """Unit test for markup save including as_df."""
    script = get_script("markup_no_results.sql")

    exported_paths_for_current_test = export_markup_combinations_from_script(
        script=script, run=False, run_args={}
    )
    paths_under_test_mapped_to_validation_paths = {
        p: get_validation_file(path1=p) for p in exported_paths_for_current_test
    }

    for (
        p_under_test,
        p_validation,
    ) in paths_under_test_mapped_to_validation_paths.items():
        assert contents_are_identical(path1=p_under_test, path2=p_validation)


@pytest.mark.markup
def test_markup_using_template_anchor_attributes(sn):
    """Unit test for markup save using an __anchor__ that is included in
    the `script.markdown.attributes.markers` section of ``snowmobile.toml``.."""
    script = get_script("markup_template_anchor.sql")

    exported_paths_for_current_test = export_markup_combinations_from_script(
        script=script, run=False, run_args={}
    )
    paths_under_test_mapped_to_validation_paths = {
        p: get_validation_file(path1=p) for p in exported_paths_for_current_test
    }

    for (
        p_under_test,
        p_validation,
    ) in paths_under_test_mapped_to_validation_paths.items():
        assert contents_are_identical(path1=p_under_test, path2=p_validation)


@pytest.mark.markup
def test_markup_scaffolding(sn, tmpdir):
    """Verifies calling markup.save() without a pre-existing directory."""
    from pathlib import Path

    # given
    script = get_script("markup_template_anchor.sql")

    # when
    script.path = Path(tmpdir) / script.path.name
    markup = script.doc()
    markup.save()

    # then
    assert not script.path.exists()  # script path is set but never written to
    assert Path(markup._path_md).exists()  # doc paths take directory of script path
    assert Path(markup._path_sql).exists()


@pytest.mark.markup
def test_markup_dunder_methods(sn):
    """Verifies __str__ and __repr__ do not cause errors."""
    script = get_script("markup_template_anchor.sql")
    markup = script.doc()
    assert str(markup)
    assert repr(markup)



"""

----

A :class:`Section` is created by a :class:`~snowmobile.core.Statement` or a
:class:`~snowmobile.core.cfg.Marker` when the :meth:`~snowmobile.Script.doc()`
method is called on an instance of :class:`~snowmobile.Script`.


.. note::

    :meth:`script.doc() <snowmobile.Script.doc()>` returns a
    :class:`~snowmobile.core.markup.Markup` instance containing a :class:`Section`
    for each statement or marker *within the script's
    context at the time* :meth:`~snowmobile.Script.doc()` *was called*.

    Respecting the script's context enables calling
    :meth:`script.doc() <snowmobile.Script.doc()>` on the instance of
    ``script`` returned by :meth:`script.filter() <snowmobile.Script.filter()>`
    and exporting an annotated markdown file and/or a tidied sql file that
    include only the contents within that context.


By default this results in two files being exported into a `.snowmobile` folder
in the same directory as the .sql file with which the instance of
:class:`~snowmobile.Script` was instantiated.

Header-levels and formatting of tagged information is configured in the
*snowmobile.toml* file, with defaults resulting in the following structure::

        ```md

        # Script Name.sql         *[script name gets an 'h1' header]
        ----

        - **Tag1**: Value1         [tags are bolded, associated values are not]
        - **Tag2**: Value2         [same for all tags/attributes found]
        - ...

        **Description**           *[except for the 'Description' section]
                                  *[this is just a blank canvas of markdown..]
                                  *[..but this is configurable]

        ## (1) create-table~dummy_name *[st get 'h2' level headers]
        ----

        - **Tag1**: Value1       *[tags can also be validations arguments..
        - **Arg1**: Val2          [that snowmobile will run on the sql results]

        **Description**          *[st get one of these too]

        **SQL**                  *[their rendered sql does as well]
            ...sql
                ...
                ...
            ...


        ## (2) update-table~dummy_name2
        ----
        [structure repeats for all statements in the script]

        ```

"""
from __future__ import annotations

import re

from typing import Dict, List, Optional, Tuple

import pandas as pd
import tabulate
tabulate.PRESERVE_WHITESPACE = True
tabulate.WIDE_CHARS_MODE = False
tabulate.MIN_PADDING = 0

from . import Generic, Configuration, errors
from .cfg import Markup
from .utils import parsing as p


class Name(Generic):
    """Handles attribute-name parsing including identification of wildcards.

    """

    def __init__(self, nm: str, cfg: Configuration, is_title: Optional[bool] = None):
        super().__init__()

        cfg_md = cfg.script.markup
        cfg_script = cfg.script

        self.nm_raw = nm

        self.nm_stripped, self.flags = cfg.wildcards.partition_on_wc(attr_nm=nm)

        self.is_paragraph = cfg_script.patterns.wildcards.wc_paragraph in self.flags
        self.is_no_reformat = cfg_script.patterns.wildcards.wc_as_is in self.flags
        self.is_omit_nm = cfg_script.patterns.wildcards.wc_omit_attr_nm in self.flags

        self.is_results = self.check_reserved_nm(
            attr_name=self.nm_stripped,
            searching_for=cfg_md.attrs.reserved["query-results"].attr_nm,
        )
        self.is_sql = self.check_reserved_nm(
            attr_name=self.nm_stripped,
            searching_for=cfg_md.attrs.reserved["rendered-sql"].attr_nm,
        )

        if self.is_omit_nm:
            self.nm_adj = str()
            self.is_paragraph = True
        elif self.is_no_reformat:
            self.nm_adj = self.nm_stripped
        elif self.nm_stripped in cfg_md.attrs.from_namespace:
            self.nm_adj = cfg_md.attrs.from_namespace.get(self.nm_stripped)
        else:
            self.nm_adj = self.nm_stripped.title() if not is_title else self.nm_stripped

        self.specified_position = cfg_md.attrs.get_position(attr=self.nm_stripped)

    @staticmethod
    def check_reserved_nm(attr_name: str, searching_for: str) -> bool:
        """Safely checks for key terms within attribute names.

        Args:
            attr_name (str):
                Attribute name that we are checking (e.g. 'Results\\*')
            searching_for (str):
                Keyword we are searching for (e.g. 'results')

        """
        attr_name, searching_for = attr_name.lower(), searching_for.lower()
        len_attr, len_kw = len(attr_name), len(searching_for)
        if len_attr >= len_kw:
            return attr_name.startswith(searching_for)
        else:
            return False

    def __repr__(self):
        return f"Name(nm='{self.nm_raw}', nm_adj='{self.nm_adj}')"

    def __str__(self):
        return f"Name(nm='{self.nm_raw}', nm_adj='{self.nm_adj}')"


# TESTS: Add tests for Item
class Item(Name):
    """Represents a piece of text/content within a header section."""

    _DELIMITER = "~"
    _INDENT_CHAR = "\t"

    def __init__(
        self,
        index: int,
        flattened_attrs: Tuple,
        cfg: Configuration,
        results: Optional[pd.DataFrame] = None,
        sql_md: Optional[str] = None,
        result_wrap: Optional[str] = None,
    ):
        self.result_wrap = result_wrap or str()
        cfg_md = cfg.script.markup
        self.is_first: bool = bool()
        self._include: bool = True

        self.cfg_md: Markup = cfg_md
        self.index = index
        self.indent, nested, in_script_value = flattened_attrs

        self._split = nested.split(self._DELIMITER)

        super().__init__(nm=self._split[-1], cfg=cfg)

        self.depth = len(self._split) - 1
        self.indent: str = self._INDENT_CHAR * self.depth
        parent = self._split[self.depth - 1]
        self.parent = Name(nm=parent, cfg=cfg)

        self._is_reserved: bool = False
        if self.is_results:
            self.value = self.as_results(results=results)
            self._is_reserved = True
            self._include = not results.empty
        elif self.is_sql:
            self.value = sql_md
            self._is_reserved = True
        else:
            self.value = in_script_value

    def is_sibling(self, other: Item) -> bool:
        return self.parent.nm_adj == other.parent.nm_adj

    def update(self, items: List[Item]):
        first_index_at_level = min({i.index for i in items if self.is_sibling(i)})
        if self.index == first_index_at_level and self.depth != 0:
            self.is_first = True
        return self

    def as_results(self, results: pd.DataFrame):
        results_cfg = self.cfg_md.attrs.reserved["query-results"]
        results_sub = results.head(results.shape[0])
        _format = results_cfg.tabulate_format or 'grid'
        if results.empty:
            return None
        if results_cfg.default_format != "markdown":
            return f"\n{results_sub.to_html(index=False)}\n"
        _w_s, _w_e = self.result_wrap, self.result_wrap
        if _w_s:
            _w_s, _w_e = f"{_w_s}\n", f"\n{_w_e}"
        _res = results_sub.to_markdown(index=False, tablefmt=_format)
        return f"\n{_w_s}{_res}{_w_e}\n"

    @property
    def _as_md_parent(self):
        parent_indent = (self.depth - 1) * self._INDENT_CHAR
        return f"{parent_indent}{self.cfg_md.bullet_char} {self.parent.nm_adj}"

    @property
    def _as_md(self):
        attr_nm = self._format_attr(attr=self.nm_adj)
        attr_value = self._format_attr(attr=self.value, is_value=True)
        if not self.is_paragraph:
            return f"{self.indent}{self.cfg_md.bullet_char} {attr_nm}: {attr_value}"
        elif attr_nm:
            return f"\n{attr_nm}\n{attr_value}"
        else:
            return f"\n{attr_value}"

    def _format_attr(self, attr: str, is_value: bool = False):
        if self._is_reserved or self.is_no_reformat or self.is_paragraph:
            return attr
        wrap_character = (
            self.cfg_md.attr_nm_wrap_char
            if not is_value
            else self.cfg_md.attr_value_wrap_char
        )
        return f"{wrap_character}{attr}{wrap_character}"

    @property
    def md(self):
        if not self.is_first:
            return self._as_md
        else:
            return f"{self._as_md_parent}\n{self._as_md}"

    def __bool__(self):
        return self._include

    def __repr__(self):
        return f"section.Item('{self.nm_adj}')"

    def __str__(self):
        return f"section.Item('{self.nm_adj}')"


# TESTS: Add tests for Section
class Section(Generic):
    """Represents any (1-6 level) header section within `Script Name (doc).md`.


    Class is created with a call to the
    :meth:`~snowmobile.core.statement.Statement.as_section()` method or
    by the :class:`snowmobile.core.markup.Markup` class in the case of a
    :class:`~snowmobile.core.cfg.script.Marker`.

    In order to include execution metadata if available without sacrificing
    base-case parsing, the below implementation heavily relies
    properties over attributes to reconcile what's populated in
    the :attr:`~snowmobile.Script.st` vs
    :attr:`~snowmobile.Script.executed` attributes of
    :class:`~snowmobile.core.Script`.

    Parameters:
        is_marker (bool):
            Information pr is associated with a marker as opposed to
            a statement; defaults to *False*.
        h_contents (str):
            String representation of header contents.
        index (int):
            Statement index position or *None* if marker.
        parsed (dict):
            Parsed arguments from the statement or marker within the script.
        raw (str):
            Raw wrap as :attr:`parsed` was parsed from.
        sql (str):
            Statement's raw sql or *None* if marker.
        results (pd.DataFrame):
            Results returned by statement's sql as a
            :class:`~pandas.DataFrame`; will be *None* if statement hasn't
            been executed or if a marker.

    Attributes:
        hx (str):
            String representation of header level (e.g. '#' for h1),
            based on the script/statement header-level specifications in
            `snowmobile.toml`.

    """

    def __init__(
        self,
        cfg: Configuration,
        is_marker: bool = None,
        h_contents: Optional[str] = None,
        index: Optional[int] = None,
        parsed: Optional[Dict] = None,
        raw: Optional[str] = None,
        sql: Optional[str] = None,
        results: Optional[pd.DataFrame] = None,
        incl_sql_tag: bool = False,
        is_multiline: bool = False,
        result_wrap: Optional[str] = None,
    ):
        """Instantiation of a ``script.Section`` object."""
        super().__init__()
        self.result_wrap = result_wrap or str()
        self.cfg: Configuration = cfg
        self.is_marker = is_marker or bool()
        self.is_multiline = is_multiline
        self.sql: str = sql
        self.transpose = parsed.get('transpose', False)
        if self.transpose:
            _ = parsed.pop('transpose')
        self.results = results if not self.transpose else results.transpose()
        if self.transpose and not self.results.empty:
            self.results.reset_index(inplace=True)
            if self.results.shape[1] == 2:
                self.results.columns = ['field name', 'field value']
        self.index: int = index
        self.raw = raw or str()
        self.incl_sql_tag = incl_sql_tag

        self._name = Name(nm=h_contents, cfg=cfg, is_title=True)
        if len(self._name.flags) > 1:
            raise errors.InvalidTagsError(
                msg=self._exception_invalid_title(raw=h_contents)
            )
        self.hx = self.cfg.markdown.pref_header(
            is_marker=self.is_marker,
            from_wc=self._name.flags[0] if self._name.flags else str(),
        )
        self.h_contents = self._name.nm_adj

        grouped_attrs = self.cfg.attrs.group_parsed_attrs(parsed)
        self.parsed: Dict = self.reorder_attrs(parsed=grouped_attrs, cfg=cfg)
        # self.items: List[Item] = [i for i in self.parse_contents(cfg=cfg) if i]
        self.items: List[Item] = self.parse_contents(cfg=cfg)

    def reorder_attrs(self, parsed: dict, cfg: Configuration) -> Dict:
        """Re-orders parsed attributes based on configuration."""
        parsed = self.cfg.attrs.add_reserved_attrs(parsed, self.is_marker)
        specified_position_to_attr_nm: Dict[int, str] = {
            Name(nm=k, cfg=cfg).specified_position: k
            for k in parsed
            if Name(nm=k, cfg=cfg).specified_position
        }
        for position in sorted(specified_position_to_attr_nm):
            attr_nm = specified_position_to_attr_nm[position]
            parsed[attr_nm] = parsed.pop(attr_nm)  # re-inserting in order
        return parsed

    def parse_contents(self, cfg: Configuration) -> List[Item]:
        """Unpacks sorted dictionary of parsed attributes into formatted Items."""
        flattened = p.dict_flatten(
            attrs=self.parsed, bullet_char=cfg.script.markup.bullet_char
        )
        items = [
            Item(
                index=i,
                flattened_attrs=v,
                cfg=cfg,
                results=self.results,
                sql_md=self.sql_md,
                result_wrap=self.result_wrap,
            )
            for i, v in enumerate(flattened, start=1)
        ]
        return [i.update(items=items) for i in items]

    @property
    def header(self) -> str:
        """Constructs the header for a section.

        Uses specifications in `snowmobile.toml` to determine:
            (1) The level of the header depending on whether it's a
                statement section or a script section.
            (2) Whether or not to include the statement index as part of the
                header.

        Returns:
            Formatted header line as a string.

        """
        if self.is_marker or not self.cfg.markdown.incl_index_in_sh:
            return f"{self.hx} {self.h_contents}"
        else:
            return f"{self.hx} ({self.index}) {self.h_contents}"

    @property
    def sql_md(self) -> str:
        """Returns renderable sql or an empty string if script-level section."""
        sql = self.cfg.script.power_strip(
            val_to_strip=("" if self.is_marker else self.sql),
            chars_to_strip=[";", " ", "\n"],
        )
        if self.incl_sql_tag:
            raw = self.cfg.script.as_parsable(self.raw, self.is_multiline)
            sql = f"{raw}\n{sql}"
        if not self.is_marker:
            sql = f"{sql};"
        if self.is_marker and not self.incl_sql_tag:
            return str()
        char = '\n'
        repl = '\t'
        return f"```sql\n{re.sub(repl, '  ', sql.strip(char))}\n```"

    @property
    def body(self) -> str:
        """All section content except for header."""
        return "\n".join(i.md for i in self.items if not isinstance(i.value, type(None)))

    @property
    def md(self) -> str:
        """Constructs a full section as a string from various components.

        Returns:
            Full string of valid markdown for the section.

        """
        markdown = "\n".join([self.header, self.body])
        return f"\n{markdown}\n"

    @staticmethod
    def _exception_invalid_title(raw: str) -> str:
        """Invalid title message; isolating to avoid cluttering __init__ method."""
        return f"""
Multiple sets of wildcards detected in the below statement or marker title; only
a single set of unescaped wildcards is permitted.
```
{raw}
```
"""

    def __repr__(self):
        return f"snowmobile.core.section.Section({str(vars(self))})"

    def __str__(self):
        return f"snowmobile.core.section.Section('{self.h_contents}')"



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['snowmobile',
 'snowmobile.core',
 'snowmobile.core.cfg',
 'snowmobile.core.pkg_data',
 'snowmobile.core.utils']

package_data = \
{'': ['*'],
 'snowmobile': ['snowmobile.egg-info/*'],
 'snowmobile.core': ['pkg_data/.snowmobile/*', 'pkg_data/.snowmobile/DDL/*']}

install_requires = \
['snowflake-connector-python==2.3.7',
 'pandas==1.1.5',
 'pydantic==1.7.2',
 'appdirs==1.4.4',
 'toml==0.10.1',
 'tabulate==0.8.7',
 'brotli==1.0.9',
 'sqlparse==0.3.1',
 'ipython==7.19.0',
 'ipykernel==5.4.2']

extras_require = \
{'docs': ['appdirs==1.4.4',
          'argon2-cffi==20.1.0',
          'asn1crypto==1.4.0',
          'astroid==2.4.2',
          'async-generator==1.10',
          'atomicwrites==1.4.0',
          'auditwheel==3.1.1',
          'azure-common==1.1.25',
          'azure-core==1.8.2',
          'azure-storage-blob==12.5.0',
          'azure-storage-common==2.1.0',
          'Babel==2.9.0',
          'backcall==0.2.0',
          'beautifulsoup4==4.9.3',
          'bleach==3.2.1',
          'boto3==1.15.18',
          'botocore==1.18.18',
          'Brotli==1.0.9',
          'CacheControl==0.12.6',
          'cachy==0.3.0',
          'certifi==2020.12.5',
          'cffi==1.14.5',
          'chardet==3.0.4',
          'cleo==0.7.6',
          'clikit==0.4.3',
          'colorama==0.4.4',
          'coverage==5.5',
          'cryptography==2.9.2',
          'css-html-js-minify==2.5.5',
          'decorator==4.4.2',
          'defusedxml==0.6.0',
          'docutils==0.15.2',
          'entrypoints==0.3',
          'flit==3.1.0',
          'flit-core==3.1.0',
          'gitdb==4.0.5',
          'GitPython==3.1.12',
          'html5lib==1.0.1',
          'idna==2.10',
          'ijson==2.6.1',
          'imagesize==1.2.0',
          'importlib-metadata==1.1.3',
          'iniconfig==1.1.1',
          'ipykernel==5.4.2',
          'ipython==7.19.0',
          'ipython-genutils==0.2.0',
          'ipywidgets==7.6.3',
          'isodate==0.6.0',
          'jedi==0.18.0',
          'Jinja2==2.11.2',
          'jmespath==0.10.0',
          'jsonschema==3.2.0',
          'jupyter-cache==0.4.2',
          'jupyter-client==6.1.11',
          'jupyter-core==4.7.0',
          'jupyter-sphinx==0.3.1',
          'jupyterlab-widgets==1.0.0',
          'keyring==20.0.1',
          'lazy-object-proxy==1.4.3',
          'lockfile==0.12.2',
          'lxml==4.6.2',
          'markdown-it-py==0.6.2',
          'MarkupSafe==1.1.1',
          'mdit-py-plugins==0.2.6',
          'mistune==0.8.4',
          'msgpack==1.0.0',
          'msrest==0.6.19',
          'myst-nb==0.12.0',
          'myst-parser==0.13.5',
          'nbclient==0.5.1',
          'nbconvert==5.6.1',
          'nbdime==2.1.0',
          'nbformat==5.1.2',
          'nest-asyncio==1.4.3',
          'notebook==6.2.0',
          'numpy==1.18.4',
          'oauthlib==3.1.0',
          'oscrypto==1.2.1',
          'packaging==20.8',
          'pandas==1.1.5',
          'pandocfilters==1.4.3',
          'parso==0.8.1',
          'pastel==0.2.0',
          'pexpect==4.8.0',
          'pickleshare==0.7.5',
          'pkginfo==1.5.0.1',
          'pluggy==0.13.1',
          'pockets==0.9.1',
          'prometheus-client==0.9.0',
          'prompt-toolkit==3.0.10',
          'ptyprocess==0.6.0',
          'py==1.10.0',
          'pyarrow==0.17.0',
          'pycparser==2.20',
          'pycryptodomex==3.9.8',
          'pydantic==1.7.2',
          'pyelftools==0.26',
          'Pygments==2.7.4',
          'PyJWT==1.7.1',
          'pylev==1.3.0',
          'pyOpenSSL==19.1.0',
          'pyparsing==2.4.7',
          'pyrsistent==0.14.11',
          'pytest==6.2.2',
          'python-dateutil==2.8.1',
          'python-slugify==4.0.1',
          'pytz==2020.1',
          'pywin32-ctypes==0.2.0',
          'pywinpty==0.5.7',
          'PyYAML==5.4',
          'pyzmq==21.0.1',
          'requests<2.24.0',
          'requests-oauthlib==1.3.0',
          'requests-toolbelt==0.8.0',
          's3transfer==0.3.3',
          'Send2Trash==1.5.0',
          'shellingham==1.3.2',
          'six==1.15.0',
          'smmap==3.0.4',
          'snowballstemmer==2.0.0',
          'snowflake-connector-python==2.3.7',
          'soupsieve==2.1',
          'Sphinx==3.5.3',
          'sphinx-autoapi==1.7.0',
          'sphinx-copybutton==0.3.1',
          'sphinx-material==0.0.32',
          'sphinx-panels==0.5.2',
          'sphinx-togglebutton==0.2.3',
          'sphinxcontrib-applehelp==1.0.2',
          'sphinxcontrib-devhelp==1.0.2',
          'sphinxcontrib-dotnetdomain==0.4',
          'sphinxcontrib-golangdomain==0.2.0.dev0',
          'sphinxcontrib-htmlhelp==1.0.3',
          'sphinxcontrib-jsmath==1.0.1',
          'sphinxcontrib-napoleon==0.7',
          'sphinxcontrib-qthelp==1.0.3',
          'sphinxcontrib-serializinghtml==1.1.4',
          'SQLAlchemy==1.3.22',
          'sqlparse==0.3.1',
          'tabulate==0.8.7',
          'terminado==0.9.2',
          'testpath==0.4.4',
          'text-unidecode==1.3',
          'toml==0.10.1',
          'tomlkit==0.5.11',
          'tornado==6.1',
          'traitlets==5.0.5',
          'typed-ast==1.4.2',
          'typing-extensions==3.7.4.3',
          'Unidecode==1.1.2',
          'urllib3==1.25.11',
          'wcwidth==0.2.5',
          'webencodings==0.5.1',
          'widgetsnbextension==3.5.1',
          'wincertstore==0.2',
          'wrapt==1.12.1',
          'zipp==3.1.0'],
 'test': ['pytest >= 5.0.0',
          'pytest-cov>=2.10.1',
          'flit>=3.0.0',
          'flit-core>=3.0.0']}

setup(name='snowmobile',
      version='0.2.0b23',
      description='An analytics-focused wrapper around the snowflake.connector for Python.',
      author='Grant Murray',
      author_email='gmurray203@gmail.com',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7.0',
     )



"""
Setting attributes on a SQL object.
../docs/snippets/set_obj_name.py
"""
import snowmobile

sn = snowmobile.connect()  # establish a connection

sn.auto_run = False  # turn off sql.auto_run

sn.query("create temp table demo_table as select 1 as sample_column;")  # demo table


try:
    sql = sn.drop()  # requires table='demo_table'
except ValueError as e:
    raise
# > ValueError: Value pr for 'table' is not valid..

try:
    sn.obj_name = "demo_table"  # set 'obj_name' attribute of SQL object
    sql = sn.drop()  # will now fallback to attribute if omitted
    print(sql)
except ValueError as e:
    raise
# stdout> 'drop table sandbox.DEMO_TABLE'

# -- complete example; should run 'as is' --



# import link
from . import link

link_name = "DictCursor"
user_text = "DictCursor"
url = "https://docs.snowflake.com/en/user-guide/python-connector-api.html#cursor"

link.xref_links.update({link_name: (user_text, url)})



"""
Demonstrate prefixing object names with an alternative schema.
../docs/snippets/sql/sql_cross_schema.py
"""
import snowmobile

sn = snowmobile.connect()

# -- SETUP --------------------------------------------------------------------
setup_sql = (
    """
    create or replace table sample_table as with
    sample_data as (
      select
        uniform(1, 10, random(1)) as rand_int
      from table(generator(rowcount => 3)) v
    )
      select
        row_number() over (order by a.rand_int) as col1
        ,(col1 * col1) as col2
      from sample_data a
    """
)

sn.ex(setup_sql)  # create 'sample_table'

# -- EXAMPLE ------------------------------------------------------------------
import snowmobile
from snowflake.connector.errors import DatabaseError

try:
    # setup
    schema_nms = ['sample_schema1', 'sample_schema2']
    for schema in schema_nms:
        _ = snowmobile.connect().ex(f"create or replace schema {schema}")

    sn = snowmobile.connect()
    assert sn.current('schema').lower() not in schema_nms

    # ==================
    # - Start Example -
    # ==================

    # Clone some tables
    sn.clone(nm='sample_table', to='other_schema.sample_table')
    sn.drop(nm='other_schema.sample_table')
    sn.clone(  # other to current schema
        nm='other_schema.sample_table',
        to='sample_table',
    )

    # Query metadata
    print(sn.exists('sample_table'))                 #> True
    print(sn.exists('sample_schema.sample_table2'))  #> True
    print(sn.exists('gem7318.sample_table3'))                #> True
    print(sn.current_schema())

    # sn.drop() works the same way
    for t in [
        'sample_table',
        'sample_schema.sample_table2',
        # 'sample_table3',
    ]:
        sn.drop(t)

    # ==================
    # - End Example -
    # ==================

except DatabaseError as e:
    raise e

finally:
    # teardown
    sn.drop(nm='sample_schema', obj='schema')
# snowmobile-include

sn = snowmobile.connect()
sn.current('schema').lower()


print(sn.select(nm='sample_table', run=False))



"""
Demonstrate core functionality of snowmobile.SQL object.
../docs/snippets/sql_working_example.py
"""
import snowmobile

from pathlib import Path

sn = snowmobile.connect()

path = Path.cwd() / "docs/snippets/dummy_table.sql"
_ = snowmobile.Script(path=path, sn=sn).run(1)

# -- 1.1 ----------------------------------------------------------------------

sample1 = sn.select("dummy_table", n=5)

print(type(sample1))  #> <class 'pandas.core.frame.DataFrame'>
print(sample1.shape)  #> (5, 3)

sample1_run_false = sn.select("dummy_table", n=5, run=False)

print(type(sample1_run_false))  #> <class 'str'>
print(sample1_run_false)

# -- 1.2 ----------------------------------------------------------------------

sn.auto_run = False

sample2 = sn.select("dummy_table", n=5)

print(type(sample2))                 #> <class 'str'>
print(sample2 == sample1_run_false)  #> True

sample2_run_true = sn.select("dummy_table", n=5, run=True)

print(type(sample2_run_true))            #> <class 'pandas.core.frame.DataFrame'>
print(sample2_run_true.equals(sample1))  #> True

# -- complete example; should run 'as is' --



"""
:class:`snowmobile.Table` is a canned implementation of
the :xref:`bulk loading from a local file system` standard and is intended to
provide a predictable, no-nonsense method of loading a :xref:`DataFrame`,
``df``, into a ``table`` (:class:`str`).

------

.. note::
   
   **Core functionality includes:**
   
   #.   Generating and executing generic DDL for ``df`` if the table doesn't yet
        exist
   
   #.   Executing DDL for the **file format** being used `if it doesn't yet
        exist in the current schema`, or (optionally) specifying an alias
        for a file format in its ``file_format`` argument; **in the case of the
        latter:**

        * An absolute ``path`` to an independent, user-defined sql file must be
          specified within the **external-sources.ddl** field of
          ``snowmobile.toml``
        * Prior to attempting the load of ``df``, :class:`snowmobile.Table`
          will create a :class:`~snowmobile.core.script.Script` from the
          configured ``path`` and execute the (file format DDL) statement whose
          tagged name maps to the value provided to its ``file_format`` argument
        * An error will be thrown during the creation of the :class:`Table` if
          the :class:`~snowmobile.core.script.Script` associated with the
          configured ``path`` does not contain a statement whose tagged name
          matches the value of ``file_format`` or if an error is raised when
          the file is parsed
        * Bypassed by creating the
          :class:`~snowmobile.Table` with:
          
          .. code:: python
          
             snowmobile.Table(validate_format=False, **kwargs)
   
   #.   Dimensional compatibility checks between ``df`` and the table being loaded
        into

        * Bypassed by creating the
          :class:`~snowmobile.Table` with:
          
          .. code:: python
          
             snowmobile.Table(validate_table=False, **kwargs)
   
   #.   Coercing column names of ``df`` into a generic database standard prior
        to loading, including de-duplication of field names when applicable
   
   #.   Argument or configuration based handling of action to take if table being
        loaded into already exists (respectively) via the ``if_exists`` argument to
        :class:`snowmobile.Table()<snowmobile.Table>` or its associated section in
        ``snowmobile.toml``; valid values are **replace**, **truncate**, **append**,
        **fail**


"""
from __future__ import annotations

import csv
import os
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd
from pandas.io.sql import DatabaseError as pdDataBaseError
from snowflake.connector.errors import DatabaseError, ProgrammingError

from . import Generic, SQL
from .connection import Snowmobile
from . import Script, errors, ExceptionHandler
from .paths import DDL_DEFAULT_PATH


class Table(Generic):
    """Constructed with a :class:`DataFrame` and a table name to load into.

    The ``df`` and ``table``'s compatibility can be inspected prior to
    calling the :meth:`Table.load()<snowmobile.table.Table.load()>` method
    or by providing `as_is=True`` when instantiating the object; the latter
    will kick off the loading process invoked by
    :meth:`.load()<snowmobile.core.table.Table.load()>`
    based on the parameters provided to
    :class:`snowmobile.Table()<snowmobile.core.table.Table>`.

    Parameters:
        df (DataFrame):
            The :class:`~pandas.DataFrame` to load.
        table (str):
            The table name to load ``df`` into.
        sn (Optional[Snowmobile]):
            An instance of :class:`~snowmobile.Snowmobile`; can be used to
            load a table on a specific connection or from a specific
            ``snowmobile.toml`` file.
        if_exists (Optional[str]):
            Action to take if ``table`` already exists - options are
            `fail`, `replace`, `append`, and `truncate`; defaults to
            ``append``.
        as_is (bool):
            Load ``df`` into ``table`` based on the parameters pr to
            :class:`Table` without further pre-inspection by the user; defaults
            to `False`.
        path_ddl (Optional[Path]):
            Alternate path to file format DDL to use for load.
        keep_local (Optional[bool]):
            Keep local file that is written out as part of the bulk loading
            process; defaults to `False`.
        path_output (Optional[str Path]):
            Path to write output local file to; defaults to a ge file
            name exported in the current working directory.
        file_format (Optional[str]):
            The name of the file_format to use when loading ``df``; defaults
            to ``snowmobile_default_psv``.
        incl_tmstmp (Optional[bool]):
            Include timestamp of load as part of ``table``; defaults to `True`.
        tmstmp_col_nm (Optional[str]):
            Name to use for load timestamp if ``incl_tmstmp=True``; defaults to
            `loaded_tmstmp`.
        upper_case_cols (Optional[bool]):
            Upper case columns of ``df`` when loading into ``table``; defaults
            to `True`.
        reformat_cols (Optional[bool]):
            Reformat applicable columns of ``df`` to be DB-compliant; defaults
            to `True`.

            Reformatting primarily entails:
                -   Replacing spaces and special characters with underscores
                -   De-duping consecutive special characters
                -   De-duping repeated column names; adds an ``_i`` suffix to
                    duplicate fields where ``i`` is the nth duplicate name for
                    a field

        validate_format (Optional[bool]):
            Validate the :xref:`file format` being used prior to kicking off
            the load; defaults to `True`.

            Validation entails:
                -   Checking if the file format being used already exists
                    based on formats accessible to the current connection
                -   Executing DDL for the file format being used if not,
                    pulled from the ``DDL`` `ext-location` and the statement
                    name ``create file format~{format name}``

            .. tip::
                 Providing `validate_format=False` will speed up loading
                 time when batch-loading into an existing table by skipping
                 this step

        validate_table (Optional[bool]):
            Perform validations of ``df`` against ``table`` prior to kicking
            off the loading process; defaults to `True`.

            Validation entails:
                -   Checking the existence of ``table``; no further
                    validation is performed if it does **not** exist
                -   Compares the columns of ``df`` to the columns of ``table``
                    and stores results for use during loading process

            .. note::
                Table validation results are used in conjunction
                with the ``if_exists`` parameter to determine the desired
                behavior based on the (potential) existence of ``table``
                and its compatibility with ``df``.

            .. tip::
                 Providing `validate_table=False` will speed up loading time
                 time when batch-loading into an existing table

        lower_case_table (Optional[bool]):
            Lower case ``table`` name; defaults to `False`.
        on_error (Optional[str]):
            Action to take if an exception is encountered as part of the
            validating or loading process - providing ``on_error='c'`` will
            *continue* past an exception as opposed to raising it; defaults to
            `None` meaning any exception encountered will be raised
        check_dupes (Optional[bool]):
            Check for duplicate field names in ``df``; defaults to `True`.
        load_copy (Optional[bool]):
            Alter and load a deep copy of ``df`` as opposed to the ``df``
            in-memory as passed to the parameter; defaults to `True`.

    Attributes:
        db_responses (Dict[str, str]):
            Responses from database during loading process.
        loaded (bool):
            Table was loaded successfully.

    """

    # noinspection PyUnresolvedReferences
    def __init__(
        self,
        df: pd.DataFrame,
        table: str,
        sn: Optional[Snowmobile] = None,
        if_exists: Optional[str] = None,
        as_is: bool = False,
        path_ddl: Optional[Union[str, Path]] = None,
        path_output: Optional[str, Path] = None,
        file_format: Optional[str] = None,
        incl_tmstmp: Optional[bool] = None,
        tmstmp_col_nm: Optional[str] = None,
        reformat_cols: Optional[bool] = None,
        validate_format: Optional[bool] = None,
        validate_table: Optional[bool] = None,
        upper_case_cols: Optional[bool] = None,
        lower_case_table: Optional[bool] = None,
        keep_local: Optional[bool] = None,
        on_error: Optional[str] = None,
        check_dupes: Optional[bool] = None,
        load_copy: Optional[bool] = None,
        **kwargs,
    ):
        # TODO: (rename) 'table' --> 'nm'
        super().__init__()
        self.sn: Snowmobile = sn or Snowmobile(delay=True, **kwargs)

        # combine kwargs + snowmobile.toml
        # --------------------------------
        if_exists = self.sn.cfg.loading.kwarg(
            arg_nm="if_exists", arg_val=if_exists, arg_typ=str
        )
        file_format = self.sn.cfg.loading.kwarg(
            arg_nm="file_format", arg_val=file_format, arg_typ=str
        )
        incl_tmstmp = self.sn.cfg.loading.kwarg(
            arg_nm="incl_tmstmp", arg_val=incl_tmstmp, arg_typ=bool
        )
        tmstmp_col_nm = (
            self.sn.cfg.loading.kwarg(
                arg_nm="tmstmp_col_nm", arg_val=tmstmp_col_nm, arg_typ=str
            )
            or "loaded_tmstmp"
        )
        reformat_cols = self.sn.cfg.loading.kwarg(
            arg_nm="reformat_cols", arg_val=reformat_cols, arg_typ=bool
        )
        validate_format = self.sn.cfg.loading.kwarg(
            arg_nm="validate_format", arg_val=validate_format, arg_typ=bool
        )
        validate_table = self.sn.cfg.loading.kwarg(
            arg_nm="validate_table", arg_val=validate_table, arg_typ=bool
        )
        upper_case_cols = self.sn.cfg.loading.kwarg(
            arg_nm="upper_case_cols", arg_val=upper_case_cols, arg_typ=bool
        )
        lower_case_table = self.sn.cfg.loading.kwarg(
            arg_nm="lower_case_table", arg_val=lower_case_table, arg_typ=bool
        )
        keep_local = self.sn.cfg.loading.kwarg(
            arg_nm="keep_local", arg_val=keep_local, arg_typ=bool
        )
        on_error = self.sn.cfg.loading.kwarg(
            arg_nm="on_error", arg_val=on_error, arg_typ=str
        )
        check_dupes = self.sn.cfg.loading.kwarg(
            arg_nm="check_dupes", arg_val=check_dupes, arg_typ=bool
        )
        copy = self.sn.cfg.loading.kwarg(arg_nm="load_copy", arg_val=load_copy, arg_typ=bool)

        # sql generation and execution
        # ----------------------------
        self.sql: SQL = SQL(nm=table, _cfg=self.sn.cfg, _query_func=self.sn.query)

        # flow control
        # ------------
        self.e = ExceptionHandler(within=self).set(ctx_id=-1)
        self.msg: str = str()
        self.requires_sql: bool = bool()

        # dataframe / table information
        # -----------------------------
        df = df.copy(deep=True) if copy else df
        self.df = df.snf.upper() if upper_case_cols else df
        self.name: str = table.upper() if not lower_case_table else table
        self._exists: bool = bool()
        self._col_diff: Dict[int, bool] = dict()

        # argument specs + self.snowmobile.toml
        # --------------------------------
        self.path_ddl = Path(str(path_ddl)) if path_ddl else DDL_DEFAULT_PATH
        self.path_output = path_output or Path.cwd() / f"{table}.csv"
        self.keep_local = keep_local
        self.on_error = on_error
        self.file_format = file_format or str()
        self.if_exists = if_exists
        self.validate_table = validate_table

        # time tracking
        # -------------
        self._upload_validation_start: int = int()
        self._upload_validation_end: int = int()
        self._load_start: int = int()
        self._load_end: int = int()

        # file format validation
        # ----------------------
        if validate_format:
            
            format_exists_in_schema = self.file_format.lower() in [
                c.lower() for c in self.sql.show('file formats', names=True)
            ]
            
            if not format_exists_in_schema:  # read from source file otherwise
                
                if not self.path_ddl.exists():
                    raise FileNotFoundError(f"`{self.path_ddl}` does not exist.")
                
                ddl = Script(sn=self.sn, path=self.path_ddl)
                st_name = f"create-file format~{self.file_format}"
                args = {  # only used if exception is thrown below
                    "nm": st_name,
                    "st": list(ddl.keys(by_index=False)),
                }
                
                try:
                    ddl.run(st_name, as_df=False)
                    
                except errors.StatementNotFoundError(**args) as e:
                    raise errors.FileFormatNameError(**args) from e

        # load timestamp
        # --------------
        if incl_tmstmp:
            col_nm = tmstmp_col_nm or "loaded_tmstmp"
            self.df.snf.add_tmstmp(
                col_nm=col_nm.upper() if upper_case_cols else col_nm.lower()
            )

        # standard column formatting
        # --------------------------
        if reformat_cols:
            self.df.snf.reformat()

        # duplicate column reformatting
        # -----------------------------
        if check_dupes and self.df.snf.has_dupes:
            self.df.snf.append_dupe_suffix()

        # other
        # -----
        self.db_responses: Dict[str, str] = dict()
        self.loaded: bool = bool()

        if as_is:
            self.load()
            
    def load(
        self,
        if_exists: Optional[str] = None,
        from_script: Path = None,
        verbose: bool = True,
        **kwargs,
    ) -> Table:
        """Loads ``df`` into ``table``.

        Args:
            if_exists (Optional[str]):
                Determines behavior to take if the table being loaded into
                already exists; defaults to **append**; options are **replace**,
                **append**, **truncate**, and **fail**
            from_script (Optional[Union[Path, str]]):
                Path to sql file containing custom DDL for ``table``; DDL is
                assumed to have a valid statement name as is parsed by
                :class:`~snowmobile.core.script.Script` and following the
                naming convention of ``create table~TABLE`` where ``TABLE`` is
                equal to the value provided to the ``table`` keyword argument
            verbose (bool):
                Verbose console output; defaults to **True**

        Returns (Table):
            The :class:`Table` after attempting load of ``df`` into
            ``table``; a successful load can be verified by inspecting
            :attr:`loaded`

        """
        if_exists = if_exists or self.if_exists
        if if_exists not in ("fail", "replace", "append", "truncate"):
            raise ValueError(
                f"Value passed to `if_exists` is not a valid argument;\n"
                f"Accepted values are: 'fail', 'replace', 'append', and 'truncate'"
            )

        # check for table existence; validate if so, all respecting `if_exists`
        if self.validate_table:
            self.validate(if_exists=if_exists)
            if self.e.seen(to_raise=True):
                if self.on_error != 'c':
                    raise self.e.get(to_raise=True, last=True)
                else:
                    return self

        try:
            self._stdout_starting(verbose)
            self.to_local()  # save to local file
            load_statements = self.load_statements(  # includes DDL if self.exists=False
                from_script=from_script
            )
            self._load_start = time.time()
            for i, s in enumerate(load_statements, start=1):
                self.db_responses[s] = self.sn.query(s).snf.to_list(n=1)  # store db responses
                self._stdout_progress(i, s, load_statements, if_exists, verbose)
            self.loaded = True

        except (ProgrammingError, pdDataBaseError, DatabaseError) as e:
            self.loaded = False
            self.e.collect(e=e)
            if self.on_error != 'c':
                raise e

        finally:
            self._load_end = time.time()
            self.sql.drop(nm=f"{self.name}_stage", obj="stage")  # drop stage
            if not self.keep_local:
                os.remove(str(self.path_output))
            self.df = self.df.snf.original  # revert to original form
            self._stdout_time(verbose=(not kwargs.get("silence") and self.loaded))

        return self

    @property
    def exists(self) -> bool:
        """Indicates if the target table exists."""
        if not self._exists:
            self._exists = self.sql.exists()
        return self._exists

    def col_diff(self, mismatched: bool = False) -> Dict[int, Tuple[str, str]]:
        """Returns diff detail of local DataFrame to in-warehouse table."""

        def fetch(idx: int, from_list: List) -> str:
            """Grab list item without throwing error if index exceeds length."""
            return str(from_list[idx]).lower() if idx <= (len(from_list) - 1) else None

        if self._col_diff:
            return self._col_diff
        if not self.exists:
            raise errors.LoadingInternalError(
                nm="Table.col_diff()", msg=f"called while `table.exists={self.exists}`."
            )

        cols_t = self.sql.columns()
        cols_df = list(self.df.columns)

        self._col_diff = {
            i + 1: (fetch(i, cols_t), fetch(i, cols_df))
            for i in range(max(len(cols_t), len(cols_df)))
        }
        return (
            self._col_diff if not mismatched
            else
            {
                k: v for k, v in self._col_diff.items()
                if v[0] != v[1]
            }
        )

    @property
    def cols_match(self) -> bool:
        """Indicates if columns match between DataFrame and table."""
        return all(d[0] == d[1] for d in self.col_diff().values())

    def load_statements(self, from_script: Path) -> List[str]:
        """Generates exhaustive list of the statements to execute for a given
        instance of loading a DataFrame."""
        load_statements = self._load_sql
        if self.requires_sql:
            load_statements.insert(0, self._load_prep_sql(from_script=from_script))
        return load_statements

    def to_local(self, quote_all: bool = True) -> None:
        """Export to local file via configuration in ``snowmobile.toml``."""
        export_options = self.sn.cfg.loading.export_options[self.file_format]
        if quote_all:
            export_options["quoting"] = csv.QUOTE_ALL
        self.df.to_csv(self.path_output, **export_options)

    @property
    def tm_load(self) -> int:
        """Seconds elapsed during loading."""
        return int(self._load_end - self._load_start)

    @property
    def tm_validate_load(self) -> int:
        """Seconds elapsed during validation."""
        return int(self._upload_validation_end - self._upload_validation_start)

    @property
    def tm_total(self) -> int:
        """Total seconds elapsed for load."""
        return self.tm_load + self.tm_validate_load

    def validate(self, if_exists: str) -> None:
        """Validates load based on current state through a variety of operations.

        Args:
            if_exists (str):
                Desired behavior if table already exists; intended to be passed
                in from :meth:`table.load()` by default.

        """
        self.e.set(ctx_id=-1)  # reset error handler to new context
        self._upload_validation_start = time.time()

        if not self.exists:  # no validation needed
            self.msg = f"{self.name} does not exist."
            self.requires_sql = "ddl"
            self._upload_validation_end = time.time()
            return

        self._col_diff = self.col_diff()
        if if_exists == "fail":
            self.msg = (
                f"`{self.name}` already exists and if_exists='fail' was "
                f"pr;\n'replace', 'append', or 'truncate' required "
                f"to continue load process with a pre-existing table."
            )
            e = errors.ExistingTableError(msg=self.msg, to_raise=True)
            self.e.collect(e)

        elif self.cols_match and if_exists == "append":
            self.msg = (
                f"`{self.name}` exists with matching columns to local "
                f"DataFrame; if_exists='{if_exists}'"
            )

        elif self.cols_match and if_exists == "replace":
            self.msg = (
                f"`{self.name}` exists with matching columns to local "
                f"DataFrame; if_exists='{if_exists}'"
            )
            self.requires_sql = "ddl"

        elif self.cols_match and if_exists == "truncate":
            self.msg = (
                f"`{self.name}` exists with matching columns to local "
                f"DataFrame; if_exists='{if_exists}'"
            )
            self.requires_sql = "truncate"

        elif not self.cols_match and if_exists != "replace":
            self.msg = (
                f"`{self.name}` columns do not equal those in the local "
                f"DataFrame and if_exists='{if_exists}' was specified.\nEither"
                f" provide if_exists='replace' to overwrite the existing table "
                f"or see `table.col_diff` to inspect the mismatched columns."
            )
            e = errors.ColumnMismatchError(msg=self.msg, to_raise=True)
            self.e.collect(e)

        elif not self.cols_match:
            self.msg = (
                f"`{self.name}` exists with mismatched columns to local "
                f"DataFrame; recreating the table with new DDL as specified "
                f"by if_exists='{if_exists}'"
            )
            self.requires_sql = "ddl"

        else:
            self.msg = (
                f"Unknown combination of arguments passed to "
                f"``loadable.to_table()``."
            )
            e = errors.LoadingInternalError(msg=self.msg, to_raise=True)
            self.e.collect(e)

        self._upload_validation_end = time.time()

    @property
    def _load_sql(self) -> List[str]:
        """Generates sql for create stage/put/copy st."""
        # fmt: off
        return [
            self.sql.create_stage(
                nm_stage=f"{self.name}_stage",
                nm_format=self.file_format,
                run=False,
            ),
            self.sql.put_file_from_stage(
                path=self.path_output,
                nm_stage=f"{self.name}_stage",
                run=False,
            ),
            self.sql.copy_into_table_from_stage(
                nm=self.name,
                nm_stage=f"{self.name}_stage",
                run=False
            ),
        ]
        # fmt: on

    def _load_prep_sql(self, from_script: Path) -> str:
        """Generates table DDL or truncate statement where applicable."""
        if self.requires_sql == "ddl" and not from_script:
            return self.df.snf.ddl(table=self.name)
        # TODO: Add 'ddl' keyword argument for the option to just pass in
        #   DDL directly as opposed to from a script.
        elif self.requires_sql == "ddl":
            return Script(path=from_script, sn=self.sn).s(_id=self.name).sql()
        else:
            return self.sql.truncate(nm=self.name, run=False)

    def _stdout_starting(self, verbose: bool):
        """Starting message."""
        if verbose:
            schema = self.sn.con.schema.lower()
            print(f"Loading into '{schema}.{self.name}`..")

    @staticmethod
    def _stdout_progress(
        i: int, s: str, st: List, if_exists: str, verbose: bool
    ) -> None:
        """"Progress message for stdout, including only first line of DDL."""
        if verbose:
            if i == 1 and len(st) == 4 and if_exists != "truncate":
                s = s.split("\n")[0] + " .."

            print(f"({i} of {len(st)})\n\t{s}")

    def _stdout_time(self, verbose: bool) -> None:
        """Time summary message for stdout."""
        if verbose:
            print(f"..completed: {self.df.shape[0]} rows in {self.tm_total} seconds")

    def __str__(self) -> str:
        return f"snowmobile.Table(table='{self.name}')"

    def __repr__(self) -> str:
        return f"snowmobile.Table(table='{self.name}')"
